import multiprocessing as mp
import time
import tkinter as tk
import traceback as tb
from idlelib.tooltip import Hovertip
from threading import Thread
from tkinter import messagebox, ttk
from typing import Optional

import numpy as np
from matplotlib.figure import Figure

import smlfm
import smlfm.graphs

from .app_model import AppModel, IStage
from .consts import COMMAND, IMAGE, TEXTVARIABLE, VARIABLE, GraphType, StageType
from .figure_window import FigureWindow
from .filter_cfg_dialog import FilterCfgDialog


# pylint: disable=too-many-ancestors
class FilterFrame(ttk.Frame, IStage):

    def __init__(self, master, model: AppModel, *args, **kwargs):
        super().__init__(master, *args, **kwargs)

        self._stage_type = StageType.FILTER
        self._stage_type_next = StageType.CORRECT

        self._model = model

        # Controls
        self._ui_frm_lbl = ttk.Label(self)
        self._ui_frm = ttk.LabelFrame(self, labelwidget=self._ui_frm_lbl)

        self._ui_summary = ttk.Frame(self._ui_frm)
        self._ui_settings = ttk.Checkbutton(self._ui_summary, style='Toolbutton')
        self._ui_settings_tip = Hovertip(self._ui_settings, text='')
        self._ui_start = ttk.Checkbutton(self._ui_summary, style='Toolbutton')
        self._ui_start_tip = Hovertip(self._ui_start, text='')
        self._ui_summary_lbl = ttk.Label(self._ui_summary)
        self._ui_preview = ttk.Checkbutton(self._ui_summary, style='Toolbutton')
        self._ui_preview_tip = Hovertip(self._ui_preview, text='')

        # Layout
        self._ui_settings.grid(column=0, row=0, sticky=tk.NW)
        self._ui_start.grid(column=1, row=0, sticky=tk.NW)
        self._ui_summary_lbl.grid(column=2, row=0, sticky=tk.EW)
        self._ui_preview.grid(column=3, row=0, sticky=tk.NW)
        self._ui_summary.columnconfigure(index=2, weight=1)

        self._ui_summary.grid(column=0, row=0, sticky=tk.EW)
        self._ui_frm.columnconfigure(index=0, weight=1)

        self._ui_frm.grid(column=0, row=0, sticky=tk.EW)
        self.columnconfigure(index=0, weight=1)

        # Padding
        self._model.grid_slaves_configure(self._ui_summary, padx=1, pady=1)
        self._model.grid_slaves_configure(self._ui_frm, padx=5, pady=(0, 5))
        self._ui_frm.configure(padding=0)

        # Texts & icons
        self._ui_frm_lbl.configure(text='Model & Filtering',
                                   image=self._model.icons.flt_stage,
                                   compound=tk.LEFT)
        self._ui_settings[IMAGE] = self._model.icons.settings
        self._ui_settings_tip.text = 'Edit filters'
        self._ui_start[IMAGE] = self._model.icons.start
        self._ui_start_tip.text = 'Proceed with filtering'
        self._ui_preview[IMAGE] = self._model.icons.preview
        self._ui_preview_tip.text = 'Preview filtered data'

        # Hooks
        self._ui_settings[COMMAND] = self._on_settings
        self._ui_start[COMMAND] = self._on_start
        self._ui_summary.bind(
            '<Configure>', lambda _e: self._ui_summary_lbl.configure(
                wraplength=self._ui_summary_lbl.winfo_width()))
        self._ui_preview[COMMAND] = self._on_preview

        # Data
        self._var_settings = tk.IntVar()
        self._ui_settings[VARIABLE] = self._var_settings
        self._var_start = tk.IntVar()
        self._ui_start[VARIABLE] = self._var_start
        self._var_summary = tk.StringVar()
        self._ui_summary_lbl[TEXTVARIABLE] = self._var_summary
        self._var_preview = tk.IntVar()
        self._ui_preview[VARIABLE] = self._var_preview

        self._update_thread: Optional[Thread] = None
        self._update_thread_err: Optional[str] = None
        self._update_thread_abort: Optional[mp.Event] = None

        self._settings_dlg: Optional[FilterCfgDialog] = None
        self._settings_apply: bool = False
        self._update_from_settings: bool = False

        # =0 ~ disabled, >0 ~ remaining flashes, None ~ infinite
        self._flash_counter = 0
        self._flash_id = None

        self.stage_ui_init()

    def stage_type(self):
        return self._stage_type

    def stage_type_next(self):
        return self._stage_type_next

    def stage_is_active(self) -> bool:
        lfm = self._model.lfm
        mla = self._model.mla
        lfl = self._model.lfl
        return lfm is not None and mla is not None and lfl is not None

    def stage_invalidate(self):
        self._flash_stop()

        if self._model.lfl is not None:
            self._model.lfl.reset_filtered_locs()

        if not self._update_from_settings:
            self._model.destroy_graph(GraphType.FILTERED)

        self._model.stage_enabled(self._stage_type, False)

        self.stage_ui_init()
        self._model.stage_invalidate(self._stage_type_next)

    def stage_start_update(self):
        self.stage_invalidate()
        self._model.stages_ui_updating(True)
        self.winfo_toplevel().configure(cursor='watch')

        self._ui_start[IMAGE] = self._model.icons.cancel
        self._ui_start_tip.text = 'Abort processing'
        self._ui_start.configure(state=tk.NORMAL)

        def _update_thread_fn():
            try:
                self._update_task()
            except BaseException as ex:
                self._update_thread_err = str(ex)
                tb.print_exception(None, ex, ex.__traceback__)

            def _update_done():
                self._ui_start[IMAGE] = self._model.icons.start
                self._ui_start_tip.text = 'Proceed with filtering'
                self._ui_start.configure(state=tk.DISABLED)

                self._update_thread = None
                self.winfo_toplevel().configure(cursor='')

                self._model.stages_ui_updating(False)

                # Reset the flag after the stage update prints it was aborted
                self._update_thread_abort = None

                if self._update_thread_err is not None:
                    err = self._update_thread_err
                    self._update_thread_err = None

                    messagebox.showerror(
                        title='Filtering Error',
                        message=f'The points cannot be filtered:\n{err}')

                    self._settings_apply = False
                    self._update_from_settings = False

                    self.stage_invalidate()
                    return

                if self._model.stage_is_active(self._stage_type_next):

                    self._show_previews(force_update=True)

                    self._model.stage_ui_init(self._stage_type_next)
                    # Next phase updates automatically or the user continues manually
                    if self._update_from_settings:
                        self._model.stage_ui_flash(self._stage_type_next)
                    else:
                        if self._model.cfg.confirm_mla_alignment:
                            self._model.stage_ui_flash(self._stage_type_next)
                        else:
                            self._model.stage_start_update(self._stage_type_next)

                self._settings_apply = False
                self._update_from_settings = False

            self._model.invoke_on_gui_thread_async(_update_done)

        self._update_thread_abort = mp.Event()
        self._update_thread = Thread(target=_update_thread_fn, daemon=True)
        self._update_thread.start()

    def stage_ui_init(self):
        self._ui_update_done()

    def stage_ui_flash(self):
        self._flash_start()

    def stage_ui_updating(self, updating: bool):
        if updating:
            self._ui_update_start()
        else:
            self._ui_update_done()

    def _ui_update_start(self):
        self._ui_settings.configure(state=tk.DISABLED)
        if self._update_from_settings:
            settings_dlg = self._settings_dlg
            if settings_dlg is not None:
                settings_dlg.enable(False)
        self._ui_start.configure(state=tk.DISABLED)
        self._ui_preview.configure(state=tk.DISABLED)

    def _ui_update_done(self):
        self._ui_update_start()

        if self.stage_is_active():
            self._model.stage_enabled(self._stage_type, True)

            self._ui_settings.configure(state=tk.NORMAL)
            if self._update_from_settings:
                settings_dlg = self._settings_dlg
                if settings_dlg is not None:
                    settings_dlg.enable(True)
            self._ui_start.configure(state=tk.NORMAL)

            if self._model.stage_is_active(self._stage_type_next):
                self._ui_preview.configure(state=tk.NORMAL)
                locs = self._model.lfl.locs_2d.shape[0]
                count = locs - self._model.lfl.filtered_locs_2d.shape[0]
                self._var_summary.set(f'Filtered out {count} localisations')
            else:
                if (self._update_thread_abort is not None
                        and self._update_thread_abort.is_set()):
                    self._var_summary.set('Filtering aborted')
                else:
                    self._var_summary.set('No filtering done')
        else:
            self._var_summary.set('No filtering done yet')

    def _flash_start(self):
        self._flash_stop()

        def _toggle_on():
            if self._flash_counter is None or self._flash_counter > 0:
                self._var_start.set(1)
                self._flash_id = self.after(500, _toggle_off)

        def _toggle_off():
            self._var_start.set(0)
            if self._flash_counter is None or self._flash_counter > 0:
                if self._flash_counter is not None:
                    self._flash_counter -= 1
                self._flash_id = self.after(500, _toggle_on)

        self._flash_counter = None
        _toggle_on()

    def _flash_stop(self):
        self._flash_counter = 0
        if self._flash_id is not None:
            self.after_cancel(self._flash_id)
        self._var_start.set(0)

    def _on_settings(self):
        cfg_dump_old = self._model.cfg.to_json()

        def _process_cb(apply: bool):
            self._settings_apply = apply
            self._update_from_settings = True
            nonlocal cfg_dump_old
            cfg_dump_new = self._model.cfg.to_json()
            if (cfg_dump_old != cfg_dump_new
                    or not self._model.stage_is_active(self._stage_type_next)):
                cfg_dump_old = cfg_dump_new
                self.stage_start_update()
            else:
                self._show_previews()
                self._settings_apply = False
                self._update_from_settings = False

        self._settings_dlg = FilterCfgDialog(self, self._model, process_cb=_process_cb)
        self._settings_dlg.show_modal()
        self._settings_dlg = None
        if not self._var_preview.get():
            self._settings_apply = False
            self._update_from_settings = False
        self._var_settings.set(0)

    def _on_start(self):
        self._var_start.set(0)

        if self._update_thread is not None:
            if self._update_thread_abort is not None:
                if not self._update_thread_abort.is_set():
                    self._ui_start.configure(state=tk.DISABLED)
                    self._var_summary.set('Aborting...')
                    self._update_thread_abort.set()
        else:
            self._flash_stop()
            self.stage_start_update()

    def _show_previews(self, force_update: bool = False):
        if not self._model.stage_is_active(self._stage_type_next):
            return

        show = self._model.cfg.show_graphs and self._model.cfg.show_all_debug_graphs

        force_show = show or self._settings_apply
        force_update = force_update and self._model.graph_exists(GraphType.FILTERED)
        if force_show:
            self._var_preview.set(1)
        if force_show or force_update:
            self._on_preview(force_update)

    def _on_preview(self, force_update: bool = False):

        def draw(f: Figure, set_size: bool = True) -> Figure:
            return smlfm.graphs.draw_locs(
                f,
                xy=self._model.lfl.filtered_locs_2d[:, 3:5],
                lens_idx=self._model.lfl.filtered_locs_2d[:, 12],
                lens_centres=((self._model.mla.lens_centres - self._model.mla.centre)
                              * self._model.lfm.mla_to_xy_scale),
                # U,V values are around MLA centre but that offset is not included
                mla_centre=np.array([0.0, 0.0]),
                set_default_size=set_size)

        wnd = self._model.graphs[GraphType.FILTERED]
        if wnd is None:
            fig = draw(Figure())
            wnd = FigureWindow(fig, master=self,
                               title='Filtered localisations')
            wnd.bind('<Map>', func=lambda _evt: self._var_preview.set(1))
            wnd.bind('<Unmap>', func=lambda _evt: self._var_preview.set(0))
            self._model.graphs[GraphType.FILTERED] = wnd
        else:
            if force_update:
                draw(wnd.figure, set_size=False)
                wnd.refresh()

        if self._var_preview.get():
            wnd.deiconify()
        else:
            wnd.withdraw()

    # Executed on thread
    def _update_task(self):
        self._model.invoke_on_gui_thread_async(
            self._var_summary.set, 'Filtering...')
        tic = time.time()

        if self._model.cfg.filter_lenses:
            self._model.lfl.filter_lenses(self._model.mla, self._model.lfm)
        if self._model.cfg.filter_rhos is not None:
            self._model.lfl.filter_rhos(self._model.cfg.filter_rhos)
        if self._model.cfg.filter_spot_sizes is not None:
            self._model.lfl.filter_spot_sizes(self._model.cfg.filter_spot_sizes)
        if self._model.cfg.filter_photons is not None:
            self._model.lfl.filter_photons(self._model.cfg.filter_photons)

        self._model.lfl.init_alpha_uv(
            self._model.cfg.alpha_model,
            self._model.lfm,
            abort_event=self._update_thread_abort,
            worker_count=self._model.cfg.max_workers)

        if self._update_thread_abort is not None:
            if self._update_thread_abort.is_set():
                self._model.lfl.reset_filtered_locs()
                return

        if self._model.cfg.log_timing:
            print(f'Filtering and setting alpha model took'
                  f' {1e3 * (time.time() - tic):.3f} ms')



import tkinter as tk
from tkinter import ttk
from typing import Callable, Optional

from .app_model import AppModel
from .consts import READONLY


# Inspired by tkinter.simpledialog.Dialog
class CfgDialog(tk.Toplevel):
    """Class to open dialogs.

    This class is intended as a base class for custom settings dialogs.
    """

    def __init__(self, parent, model: AppModel, title: Optional[str] = None,
                 process_cb: Optional[Callable[[bool], None]] = None):
        """Initialize a dialog.

        Args:
            parent: A parent window (the application window).
            model (AppModel): A container with application data.
            title (str): The dialog title.
            process_cb (Callable[[bool], None]):
                A callback function invoked on OK and Apply button click.
                The argument is True for Apply and False for OK button.
        """

        super().__init__(parent)

        self.parent = parent
        self.model = model
        self._process_cb = process_cb
        self._process_cb_apply = False
        # UI controls to be enabled or disabled during update via 'enable' function
        self.ui_widgets = []

        self.withdraw()  # Remain invisible until 'show_modal' call
        # If the parent is not viewable, don't make the child transient,
        # or else it would be opened withdrawn.
        if parent is not None and parent.winfo_viewable():
            self.transient(parent)

        if title:
            self.title(title)

        _setup_dialog(self)

        body_box = ttk.Frame(self)
        self.initial_focus = self.body(body_box)
        if not self.initial_focus:
            self.initial_focus = self
        body_box.pack(anchor=tk.NW, fill=tk.X, expand=True, padx=5, pady=5)

        separator = ttk.Separator(self, orient='horizontal')
        separator.pack(anchor=tk.S, fill=tk.X, padx=0, pady=0)

        button_box = ttk.Frame(self)
        self.buttonbox(button_box)
        button_box.pack(anchor=tk.S, fill=tk.X, padx=5, pady=5)

        self.protocol("WM_DELETE_WINDOW", self.cancel)

        # Set min. size
        self.update()  # Geometry is '1x1+0+0' here, update it first
        self.minsize(self.winfo_width(), self.winfo_height())

        # Set default size (enlarges to min. size if necessary)
        if parent is None:
            self.geometry('450x50')
        else:
            self.geometry(f'450x50'
                          f'+{parent.winfo_rootx() + 50}'
                          f'+{parent.winfo_rooty() + 50}')

    def destroy(self):
        """Destroy the window."""
        self.initial_focus = None
        self.ui_widgets = []
        super().destroy()

    # pylint: disable=unused-argument
    def body(self, master) -> tk.BaseWidget:
        """Create dialog body.

        Return widget that should have initial focus.
        This method should be overridden, and is called
        by the __init__ method.
        """

        return self  # Override

    def buttonbox(self, master) -> None:
        """Add standard button box.

        Override if you don't want the standard buttons.

        """

        btn_ok = ttk.Button(master, text='OK', compound=tk.LEFT,
                            image=self.model.icons.ok,
                            command=self.ok, default=tk.ACTIVE)
        btn_ok.pack(side=tk.RIGHT, padx=5, pady=5)
        self.ui_widgets.append(btn_ok)

        btn_apply = ttk.Button(master, text='Apply', compound=tk.LEFT,
                               image=self.model.icons.start,
                               command=self.apply)
        btn_apply.pack(side=tk.RIGHT, padx=(5, 0), pady=5)
        self.ui_widgets.append(btn_apply)

        btn_cancel = ttk.Button(master, text='Close', compound=tk.LEFT,
                                image=self.model.icons.cancel,
                                command=self.cancel)
        btn_cancel.pack(side=tk.RIGHT, padx=(5, 0), pady=5)
        # Close/Cancel button is not disabled, allows closing dialog during update
        # self.ui_widgets.append(btn_cancel)

        self.bind("<Return>", self.ok)
        self.bind("<Escape>", self.cancel)

    def show_modal(self) -> None:
        self.deiconify()  # Become visible now
        self.enable(True)
        self.initial_focus.focus_set()

        # Wait for window to appear on screen before calling grab_set
        self.wait_visibility()
        self.grab_set()
        self.wait_window(self)

    def ok(self, _event=None) -> None:
        if not self.validate():
            self.initial_focus.focus_set()  # Put focus back
            return

        self.withdraw()
        self.update_idletasks()

        try:
            self.process()
        finally:
            self.cancel()

    def apply(self, _event=None) -> None:
        self._process_cb_apply = True
        try:
            if not self.validate():
                self.initial_focus.focus_set()  # Put focus back
            else:
                self.process()
        finally:
            self._process_cb_apply = False

    def cancel(self, _event=None) -> None:
        if self.parent is not None:
            self.parent.focus_set()  # Put focus back to the parent window
        self.destroy()

    def validate(self) -> bool:
        """Validate the data.

        This method is called automatically to validate the data before the
        dialog is destroyed. By default, it always validates OK.
        """

        return True  # Override

    def process(self) -> None:
        """Process the data.

        This method is called automatically to process the data
        on Apply and OK button click.
        By default, it invokes process callback only.
        """

        if self._process_cb is not None:
            self._process_cb(self._process_cb_apply)

        # Override

    def enable(self, active: bool, change_cursor: bool = True) -> None:
        if change_cursor:
            self.winfo_toplevel().configure(cursor='' if active else 'watch')

        for w in self.ui_widgets:
            if active:
                state = tk.NORMAL
                if isinstance(w, ttk.Combobox):
                    state = READONLY
            else:
                state = tk.DISABLED
            w.configure(state=state)

    @staticmethod
    def is_float(num) -> bool:
        try:
            float(num)
            return True
        except ValueError:
            return False

    @staticmethod
    def is_int(num) -> bool:
        try:
            int(num)
            return True
        except ValueError:
            return False

    @staticmethod
    def clamp(n, min_n, max_n):
        return min(max(n, min_n), max_n)


# noinspection PyProtectedMember
# pylint: disable=protected-access
def _setup_dialog(w) -> None:
    if w._windowingsystem == "aqua":
        w.tk.call("::tk::unsupported::MacWindowStyle", "style",
                  w, "moveableModal", "")
    elif w._windowingsystem == "x11":
        w.wm_attributes("-type", "dialog")



from pathlib import Path
from typing import Optional

_HAS_IMAGEJ: bool = True
try:
    import imagej
    import jpype
    import scyjava
except ImportError:
    _HAS_IMAGEJ = False


class FijiApp:

    def __init__(self):
        # One time detection during app start if PyImageJ package is installed
        self.has_imagej: bool = _HAS_IMAGEJ
        # Detected during update of CFG stage
        # Once detected, it cannot be deactivated as Java can only be started once
        self.has_fiji: bool = False
        # A path where Fiji app has been detected
        # Modified value will apply only after app restarts
        self.fiji_dir: Optional[Path] = None
        # A list with options for JVM
        # Modified value will apply only after app restarts
        self.fiji_jvm_opts: str = ''
        # ImageJ/Fiji instance, if detected
        if _HAS_IMAGEJ:
            self.ij: Optional[jpype.JClass] = None

    # Can be executed on thread
    def detect_fiji(self, fiji_dir: Path, fiji_jvm_opts: str = '') -> Optional[bool]:
        """ Returns None if Fiji has been already detected and this call
            was done with different path or JVM options. """
        if self.has_fiji:
            # Once detected, it cannot be deactivated as Java can only be started once
            if fiji_dir != self.fiji_dir or fiji_jvm_opts != self.fiji_jvm_opts:
                return None
            return True
        if not self.has_imagej:
            # PyImageJ not installed
            return False
        if fiji_dir is None:
            return False

        try:
            jvm_opts = fiji_jvm_opts.split()
            scyjava.config.add_options(jvm_opts)
            self.ij = imagej.init(str(fiji_dir), mode=imagej.Mode.INTERACTIVE)
        except RuntimeError:
            return False

        print(f'ImageJ version(s): {self.ij.getVersion()}')  # pylint: disable=no-member
        print(f'IJ app version: {self.ij.app().getVersion()}')  # pylint: disable=no-member

        self.has_fiji = True
        self.fiji_dir = fiji_dir
        self.fiji_jvm_opts = fiji_jvm_opts

        return True

    # Can be executed on thread
    def run_peakfit(self, img_stack: Path, out_csv_file: Path):
        j_stack = self.ij.io().open(str(img_stack))  # pylint: disable=no-member
        j_imp_stack = self.ij.py.to_imageplus(j_stack)  # pylint: disable=no-member

        res_dir = out_csv_file.parent

        # noinspection PyPep8Naming
        # pylint: disable=invalid-name
        HOME = Path.home()

        # Settings for a PeakFit ImageJ plugin
        peakfit_args = {
            'template': '',
        }
        if not (HOME / '.gdsc.smlm' / 'calibration.settings').exists():
            peakfit_args.update({
                'camera_type': 'EMCCD',
                'camera_bias': 400.00,
                'gain': 40.0000,
                'read_noise': 7.1600,
                'calibration': 266.0,
                'exposure_time': 30.0,
            })
        if not (HOME / '.gdsc.smlm' / 'psf.settings').exists():
            peakfit_args.update({
                'psf': 'Circular Gaussian 2D',
                'psf_parameter_1': 1.700,
            })
        peakfit_args.update({
            'spot_filter_type': 'Difference',
            'spot_filter2': 'Mean',
            'smoothing2': 2.0,
        })
        if not (HOME / '.gdsc.smlm' / 'fitenginesettings.settings').exists():
            peakfit_args.update({
                # 'spot_filter_type': 'Difference',
                # 'spot_filter2': 'Mean',
                # 'smoothing2': 2.0,
                'spot_filter': 'Mean',
                'smoothing': 0.32,
                'search_width': 0.90,
                'border_width': 4.00,
                'fitting_width': 4.00,

                'fit_solver': 'LVM LSE',
                'relative_threshold': 1.0E-6,
                'absolute_threshold': 1.0E-16,
                'parameter_relative_threshold': 0.0,
                'parameter_absolute_threshold': 0.0,
                'max_iterations': 20,
                'lambda': 10.0000,
                'fail_limit': 3,
                'pass_rate': 0.50,

                'neighbour_height': 0.30,
                'residuals_threshold': 1.00,
                'duplicate_distance': 0.50,
                'duplicate_distance_absolute': True,

                'shift_factor': 1.2,
                'signal_strength': 3.0,
                'min_photons': 10.0,
                'min_width_factor': 0.4,
                'max_width_factor': 1.0,
                'precision': 75.0,
                'precision_method': 'Mortensen',
            })
        # Overwrite result settings unconditionally
        # if not (HOME / '.gdsc.smlm' / 'resultssettings.settings').exists():
        peakfit_args.update({
            'results_format': 'Text',
            'file_distance_unit': 'unknown (na)',
            'file_intensity_unit': 'unknown (na)',
            'file_angle_unit': 'unknown (na)',
            'file_show_precision': True,
            'results_directory': str(res_dir),
        })
        # Temporary settings for development
        peakfit_args.update({
            # 'log_progress': True,

            'image': 'None',
            # 'image': 'Localisations (width=precision)',
            'equalised': True,
            'image_size_mode': 'Scaled',
            'image_scale': 1,
            'lut': 'Fire',
        })

        # pylint: disable=no-member
        was_log_window_open = self.ij.WindowManager.getWindow("Log") is not None

        # Apply our configuration # pylint: disable=no-member
        self.ij.py.run_plugin('Fit Configuration', args=peakfit_args)
        # Allow user to change the configuration # pylint: disable=no-member
        self.ij.py.run_plugin('Fit Configuration')
        # Run the plugin code on given stack # pylint: disable=no-member
        self.ij.py.run_plugin('Peak Fit', args=peakfit_args, imp=j_imp_stack)

        # Close Log window if it wasn't open before Peak Fit
        if not was_log_window_open:
            # pylint: disable=no-member
            log_window = self.ij.WindowManager.getWindow("Log")
            if log_window is not None:
                log_window.setVisible(False)

        res_xls_file = res_dir / (img_stack.name + '.results.xls')
        # Rename to configured CSV file (overwrite if exists)
        res_xls_file.replace(out_csv_file)



from .config import Config
from .fiji_app import FijiApp
from .fitting import Fitting
from .fourier_microscope import FourierMicroscope
from .localisation_file import LocalisationFile
from .localisations import Localisations
from .micro_lens_array import MicroLensArray
from .output_files import OutputFiles

__all__ = [
    "Config",
    "FijiApp",
    "Fitting",
    "FourierMicroscope",
    "LocalisationFile",
    "Localisations",
    "MicroLensArray",
    "OutputFiles",
]



import sys
import threading
import tkinter as tk
from pathlib import Path
from queue import Empty, SimpleQueue
from tkinter import ttk
from typing import Dict, Optional

import numpy.typing as npt

import smlfm

from .assets import Icons
from .consts import GraphType, StageType
from .figure_window import FigureWindow


class IStage:
    def stage_type(self) -> StageType:
        """Returns the stage type."""
        raise NotImplementedError('Subclass must implement this')

    def stage_type_next(self) -> Optional[StageType]:
        """Returns the next stage type."""
        raise NotImplementedError('Subclass must implement this')

    def stage_is_active(self) -> bool:
        """Returns the stage state, True when the stage has valid input."""
        raise NotImplementedError('Subclass must implement this')

    def stage_invalidate(self):
        """Invalidates the stage data."""
        raise NotImplementedError('Subclass must implement this')

    def stage_start_update(self):
        """Starts the stage data update, load resources, etc."""
        raise NotImplementedError('Subclass must implement this')

    def stage_ui_init(self):
        """Initializes the stage UI, possibly with date from previous stages."""
        raise NotImplementedError('Subclass must implement this')

    def stage_ui_flash(self):
        """Flashes with a control where the user should continue, no-op by default."""

    def stage_ui_updating(self, updating: bool):
        """Changes the stage UI state. If `updating` is True the UI is mostly
        disabled, except some 'Cancel' or 'Abort' button if exists.
        If the update is not in progress, the UI reflects internal state and
        enables UI controls that have valid input data from model."""
        raise NotImplementedError('Subclass must implement this')


# pylint: disable=too-many-instance-attributes
class AppModel:
    """A class holding application data and guards top-level logic.

    The UI and logic is split into multiple stages that build a chain.
    Invalidating first stage invalidates all other stages.
    Invalidating last stage invalidates last stage only.

    Most of the stages have input data stored in the 'cfg' member, some of them
    can visualize the output in a graph.

    The stages have the following order:
        1. CONFIG stage - Configuration
            - Invalidate:
                - Deletes `cfg`
                - Invalidates CSV stage
            - Update:
                - Creates `cfg` (from JSON file)
                - Updates CSV stage
        2. CSV stage - Localisations
            - Invalidate:
                - Deletes `csv`
                - Deletes CSV_RAW graph
                - Invalidates OPTICS stage
            - Update:
                - Creates `csv` (from CSV file and `cfg`)
                - Updates CSV_RAW graph data
                - Updates OPTICS stage
        3. OPTICS stage - Micro-Lens Array & Microscope
            - Invalidate:
                - Deletes `mla`, `lfm` and `lfl`
                - Deletes MLA and MAPPED graphs
                - Invalidates FILTER stage
            - Update:
                - Creates `lfm` and `mla` (from `cfg`)
                - Updates `csv` data scale (from `lfm`)
                - Creates `lfl` (from `lfm`, `mla` and `csv`)
                - Updates MLA and MAPPED graph data
                - Updates FILTER stage
                - Forces alignment confirmation
        4. FILTER stage - Filter mapped data
            - Invalidate:
                - Resets `lfl` filters
                - Deletes FILTERED graph
                - Invalidates CORRECT stage
            - Update:
                - Applies `lfl` filters (from UI lists)
                - Initializes `lfl.alpha_model` (from `lfl.filtered_locs_2d`)
                - Updates FILTERED graph data
                - Forces update of CORRECT stage
        5. CORRECT stage - Aberration correction
            - Invalidate:
                - Resets `lfl` corrections
                - Deletes CORRECTED graph
                - Invalidates FIT stage
            - Update:
                - Calculates `lfl` correction (from `lfl.filtered_locs_2d`)
                - Initializes `lfl.corrected_locs_2d` (from `lfl.filtered_locs_2d`)
                - Updates CORRECTED graph data
                - Forces update of FIT stage
        6. FIT stage - 3D fitting
            - Invalidate:
                - Deletes `locs_3d`
                - Deletes LOCS_3D, HISTOGRAM and OCCURRENCES graphs
            - Update:
                - Update timestamp
                - Generates `locs_3d` (from `cfg`, `lfm` and `lfl.corrected_locs_2d`)
                - Updates LOCS_3D, HISTOGRAM and OCCURRENCES graphs data
                - Saves results (from `cfg` and `locs_3d`)
    """

    _INVOKE_ON_GUI_EVENT = '<<_InvokeOnGui>>'

    # An entry put on queue from any but GUI thread.
    # The function is invoked from a GUI thread with given arguments.
    class _InvokeOnGuiData:
        def __init__(self, fn, args, kwargs):
            self.fn = fn
            self.args = args
            self.kwargs = kwargs
            self.reply = None
            self.reply_event = threading.Event()

    def __init__(self):
        self.icons = Icons()

        self._root = None
        self._min_root_width: Optional[int] = None
        self._gui_queue: Optional[SimpleQueue] = None

        self.fiji_app = smlfm.FijiApp()

        self.save_timestamp = None

        self.cli_cfg_file: Optional[Path] = None
        self.cli_img_stack: Optional[Path] = None
        self.cli_csv_file: Optional[Path] = None

        self.cfg_file: Optional[Path] = None
        self.cfg: Optional[smlfm.Config] = None

        self.csv: Optional[smlfm.LocalisationFile] = None
        self.lfm: Optional[smlfm.FourierMicroscope] = None
        self.mla: Optional[smlfm.MicroLensArray] = None
        self.lfl: Optional[smlfm.Localisations] = None

        self.locs_3d: Optional[npt.NDArray[float]] = None

        self.graphs: Dict[GraphType, Optional[FigureWindow]] = {}
        for t in GraphType:
            self.graphs[t] = None

        self._ui_mains: Dict[StageType, Optional[ttk.Frame]] = {}
        self._ui_numbers: Dict[StageType, Optional[ttk.Label]] = {}
        self._ui_frames: Dict[StageType, Optional[ttk.Frame]] = {}
        self._stages: Dict[StageType, Optional[IStage]] = {}
        for t in StageType:
            self._ui_numbers[t] = None
            self._ui_frames[t] = None
            self._stages[t] = None

        # Handle CLI options
        for arg in sys.argv[1:]:
            a = arg.casefold()
            if a.endswith('.json'):
                self.cli_cfg_file = Path(arg)
            elif a.endswith('.tiff') or a.endswith('.tif'):
                self.cli_img_stack = Path(arg)
            elif a.endswith('.csv') or a.endswith('.xls'):
                self.cli_csv_file = Path(arg)
            else:
                print(f'WARNING: Ignoring unrecognized option "{arg}"')

    def create_gui(self, root):
        # Import here to avoid circular dependency
        from .config_frame import ConfigFrame  # pylint: disable=import-outside-toplevel
        from .optics_frame import OpticsFrame  # pylint: disable=import-outside-toplevel
        from .csv_frame import CsvFrame  # pylint: disable=import-outside-toplevel
        from .filter_frame import FilterFrame  # pylint: disable=import-outside-toplevel
        from .correct_frame import CorrectFrame  # pylint: disable=import-outside-toplevel
        from .fit_frame import FitFrame  # pylint: disable=import-outside-toplevel

        self._root = root
        self._root.bind(AppModel._INVOKE_ON_GUI_EVENT, self._gui_invoke_handler)
        self._gui_queue = SimpleQueue()

        rows = {
            StageType.CONFIG: (ConfigFrame, self.icons.nr1_large),
            StageType.CSV: (CsvFrame, self.icons.nr2_large),
            StageType.OPTICS: (OpticsFrame, self.icons.nr3_large),
            StageType.FILTER: (FilterFrame, self.icons.nr4_large),
            StageType.CORRECT: (CorrectFrame, self.icons.nr5_large),
            StageType.FIT: (FitFrame, self.icons.nr6_large),
        }
        for i, t in enumerate(StageType):
            self._ui_mains[t] = ttk.Frame(self._root)
            self._ui_numbers[t] = ttk.Label(self._ui_mains[t], image=rows[t][1])
            self._ui_frames[t] = rows[t][0](self._ui_mains[t], self)

            self._ui_numbers[t].grid(column=0, row=0, sticky=tk.NSEW)
            self._ui_frames[t].grid(column=1, row=0, sticky=tk.EW)
            self._ui_mains[t].columnconfigure(index=1, weight=1)
            self._ui_mains[t].grid(sticky=tk.NSEW)

            self._ui_mains[t].configure(padding=0)
            self._ui_numbers[t].configure(padding=(0, 10 if i == 0 else 0, 0, 0))
            self._ui_frames[t].configure(padding=(0, 10 if i == 0 else 0, 10, 10))

            self._stages[t] = self._ui_frames[t]

        self._root.columnconfigure(index=0, weight=1)

        def _update_min_height(_evt=None):
            self._root.update()
            self._root.minsize(self._min_root_width, self._root.bbox()[3])
        for i, t in enumerate(StageType):
            self._ui_frames[t].bind('<Configure>', _update_min_height)

        # Set min. size
        self._root.update()  # Geometry is '1x1+0+0' here, update it first
        self._min_root_width = self._root.winfo_width()
        self._root.minsize(self._min_root_width, self._root.winfo_height())

    def invoke_on_gui_thread_async(self, fn, *args, **kwargs):
        data = AppModel._InvokeOnGuiData(fn, args, kwargs)
        if threading.main_thread().ident == threading.get_ident():
            # Call directly on GUI thread
            data.reply = data.fn(*data.args, *data.kwargs)
        else:
            self._gui_queue.put(data)
            self._root.event_generate(AppModel._INVOKE_ON_GUI_EVENT, when='tail')
            data.reply = None
        return data.reply

    def invoke_on_gui_thread_sync(self, fn, *args, **kwargs):
        data = AppModel._InvokeOnGuiData(fn, args, kwargs)
        if threading.main_thread().ident == threading.get_ident():
            # Call directly on GUI thread
            data.reply = data.fn(*data.args, *data.kwargs)
        else:
            self._gui_queue.put(data)
            self._root.event_generate(AppModel._INVOKE_ON_GUI_EVENT, when='tail')
            data.reply_event.wait()
        return data.reply

    def _gui_invoke_handler(self, _event):
        try:
            while True:
                data = self._gui_queue.get_nowait()
                data.reply = data.fn(*data.args, *data.kwargs)
                data.reply_event.set()
        except Empty:
            pass

    def stage_enabled(self, st: Optional[StageType], enabled: bool):
        if st is not None:
            self._ui_numbers[st].configure(state=tk.NORMAL if enabled else tk.DISABLED)

    def stage_is_active(self, st: Optional[StageType]):
        if st is not None:
            return self._stages[st].stage_is_active()
        return False

    def stage_invalidate(self, st: Optional[StageType]):
        if st is not None:
            self._stages[st].stage_invalidate()

    def stage_start_update(self, st: Optional[StageType]):
        if st is not None:
            self._stages[st].stage_start_update()

    def stage_ui_init(self, st: Optional[StageType]):
        if st is not None:
            self._stages[st].stage_ui_init()

    def stage_ui_flash(self, st: Optional[StageType]):
        if st is not None:
            self._stages[st].stage_ui_flash()

    def stage_ui_updating(self, st: Optional[StageType], updating: bool):
        if st is not None:
            self._stages[st].stage_ui_updating(updating)

    def stages_ui_updating(self, updating: bool):
        for st in StageType:
            self.stage_ui_updating(st, updating)

    def graph_exists(self, gt: GraphType) -> bool:
        return self.graphs[gt] is not None

    def destroy_graph(self, gt: GraphType):
        wnd = self.graphs[gt]
        self.graphs[gt] = None
        if wnd is not None:
            wnd.withdraw()  # Keep that hiding for Unmap binding to work
            wnd.destroy()

    @staticmethod
    def grid_slaves_configure(container: tk.Widget, *args, **kwargs):
        for child in container.grid_slaves():
            child.grid(*args, **kwargs)



import pkgutil
import tkinter as tk

import cairosvg

from . import __name__ as _pkg_name_


# pylint: disable=too-many-instance-attributes
class Icons:
    # pylint: disable=too-many-statements
    def __init__(self):
        # TODO: Detect high-DPI screen and adjust the scale factor
        self.svg_scale = 1.0
        self.big_number_scale = 3.0
        self._svg_big_number_scale = self.svg_scale * self.big_number_scale

        # Black filled Material icons taken from:
        # https://fonts.google.com/icons?icon.set=Material+Icons&icon.style=Filled
        self.default_size = 18  # In PX for image size, in DP for file name
        color = 'black'
        fill_hex_color = '#000000'
        mi_suffix = f'_{color}_{self.default_size}dp.svg'

        # Remaining icons taken from: https://pictogrammers.com/library/mdi/
        # These icons don't contain width and height attributes in pixels,
        # and fill value. We inject it here.
        pg_suffix = '.svg'

        def pg_inject_missing_attrs(svg_data: bytes) -> bytes:
            for attr in [b' height=', b' width=', b' fill=']:
                if attr in svg_data:
                    return svg_data  # Don't modify
            vb = 'viewBox="0 0 24 24"'
            h = f'height="{self.default_size}px"'
            w = f'width="{self.default_size}px"'
            f = f'fill="{fill_hex_color}"'
            svg_data = svg_data.replace(f' {vb}'.encode('ascii'),
                                        f' {h} {vb} {w} {f}'.encode('ascii'),
                                        1)
            return svg_data

        # This hack crops the number icons and enforces their real minimal size
        def pg_crop_number_icon(svg_data: bytes) -> bytes:
            for attr in [b' viewBox=', b' height=', b' width=']:
                if attr not in svg_data:
                    return svg_data  # Don't modify
            vb1 = 'viewBox="0 0 24 24"'
            vb2 = 'viewBox="6 6 12 12"'
            h1 = f'height="{self.default_size}px"'
            w1 = f'width="{self.default_size}px"'
            h2 = 'height="12px"'
            w2 = 'width="12px"'
            svg_data = svg_data.replace(f' {vb1}'.encode('ascii'),
                                        f' {vb2}'.encode('ascii'),
                                        1)
            svg_data = svg_data.replace(f' {h1}'.encode('ascii'),
                                        f' {h2}'.encode('ascii'),
                                        1)
            svg_data = svg_data.replace(f' {w1}'.encode('ascii'),
                                        f' {w2}'.encode('ascii'),
                                        1)
            return svg_data

        # Color palette for icons generated at:
        # https://coolors.co/d19c1d-613a3a-9d75cb-118ab2-26c485-db2955
        color_app = '#D19C1D'  # Goldenrod (yellow-like)
        color_any = '#613A3A'  # Rose ebony (brown-like)
        color_settings = '#9D75CB'  # Amethyst (violet-like)
        color_preview = '#118AB2'  # Blue (NCS)
        color_start = '#26C485'  # Emerald (green-like)
        color_cancel = '#DB2955'  # Raspberry (red-like)

        def replace_color(svg_data: bytes, new_hex_color: str,
                          old_hex_color: str = fill_hex_color) -> bytes:
            svg_data = svg_data.replace(f'fill="{old_hex_color}"'.encode('ascii'),
                                        f'fill="{new_hex_color}"'.encode('ascii'),
                                        1)
            return svg_data

        def load_svg_image(svg_data: bytes, scale: float = 1.0):
            png_data = cairosvg.svg2png(bytestring=svg_data, scale=scale)
            return tk.PhotoImage(data=png_data)

        # Material icons

        data = pkgutil.get_data(_pkg_name_, f'data/teledyne_rhomboid{mi_suffix}')
        # Keep black or change to Teledyne Blue (#0076C0)
        # data = replace_color(data, '#0076C0')
        self.app = load_svg_image(data, self.svg_scale * 32.0 / self.default_size)

        data = pkgutil.get_data(_pkg_name_, f'data/folder_open{mi_suffix}')
        data = replace_color(data, color_any)
        self.open = load_svg_image(data, self.svg_scale)

        data = pkgutil.get_data(_pkg_name_, f'data/save{mi_suffix}')
        data = replace_color(data, color_any)
        self.save = load_svg_image(data, self.svg_scale)

        data = pkgutil.get_data(_pkg_name_, f'data/save_as{mi_suffix}')
        data = replace_color(data, color_any)
        self.save_as = load_svg_image(data, self.svg_scale)

        data = pkgutil.get_data(_pkg_name_, f'data/settings{mi_suffix}')
        data = replace_color(data, color_settings)
        self.settings = load_svg_image(data, self.svg_scale)

        data = pkgutil.get_data(_pkg_name_, f'data/play_arrow{mi_suffix}')
        data = replace_color(data, color_start)
        self.start = load_svg_image(data, self.svg_scale)

        data = pkgutil.get_data(_pkg_name_, f'data/cancel{mi_suffix}')
        data = replace_color(data, color_cancel)
        self.cancel = load_svg_image(data, self.svg_scale)

        data = pkgutil.get_data(_pkg_name_, f'data/done{mi_suffix}')
        data = replace_color(data, color_start)
        self.ok = load_svg_image(data, self.svg_scale)

        data = pkgutil.get_data(_pkg_name_, f'data/visibility{mi_suffix}')
        data = replace_color(data, color_preview)
        self.preview = load_svg_image(data, self.svg_scale)

        data = pkgutil.get_data(_pkg_name_, f'data/construction{mi_suffix}')
        data = replace_color(data, color_any)
        self.cfg_stage = load_svg_image(data, self.svg_scale)

        data = pkgutil.get_data(_pkg_name_, f'data/scatter_plot{mi_suffix}')
        data = replace_color(data, color_any)
        self.csv_stage = load_svg_image(data, self.svg_scale)

        data = pkgutil.get_data(_pkg_name_, f'data/table_rows{mi_suffix}')
        data = replace_color(data, color_any)
        self.csv_data = load_svg_image(data, self.svg_scale)

        data = pkgutil.get_data(_pkg_name_, f'data/collections{mi_suffix}')
        data = replace_color(data, color_any)
        self.csv_stack = load_svg_image(data, self.svg_scale)

        data = pkgutil.get_data(_pkg_name_, f'data/search{mi_suffix}')
        data = replace_color(data, color_any)
        self.csv_peakfit = load_svg_image(data, self.svg_scale)

        data = pkgutil.get_data(_pkg_name_, f'data/biotech{mi_suffix}')
        data = replace_color(data, color_any)
        self.opt_stage = load_svg_image(data, self.svg_scale)

        data = pkgutil.get_data(_pkg_name_, f'data/filter_alt{mi_suffix}')
        data = replace_color(data, color_any)
        self.flt_stage = load_svg_image(data, self.svg_scale)

        data = pkgutil.get_data(_pkg_name_, f'data/panorama_horizontal_select{mi_suffix}')
        data = replace_color(data, color_any)
        self.cor_stage = load_svg_image(data, self.svg_scale)

        data = pkgutil.get_data(_pkg_name_, f'data/mediation{mi_suffix}')
        data = replace_color(data, color_any)
        self.fit_stage = load_svg_image(data, self.svg_scale)

        data = pkgutil.get_data(_pkg_name_, f'data/bar_chart{mi_suffix}')
        data = replace_color(data, color_preview)
        self.fit_occurrences = load_svg_image(data, self.svg_scale)

        data = pkgutil.get_data(_pkg_name_, f'data/apps{mi_suffix}')
        data = replace_color(data, color_preview)
        self.fit_histogram = load_svg_image(data, self.svg_scale)

        # Pictogrammers icons

        data = pkgutil.get_data(_pkg_name_, f'data/numeric-1{pg_suffix}')
        data = pg_inject_missing_attrs(data)
        data = pg_crop_number_icon(data)
        data = replace_color(data, color_app)
        self.nr1_large = load_svg_image(data, self._svg_big_number_scale)

        data = pkgutil.get_data(_pkg_name_, f'data/numeric-2{pg_suffix}')
        data = pg_inject_missing_attrs(data)
        data = pg_crop_number_icon(data)
        data = replace_color(data, color_app)
        self.nr2_large = load_svg_image(data, self._svg_big_number_scale)

        data = pkgutil.get_data(_pkg_name_, f'data/numeric-3{pg_suffix}')
        data = pg_inject_missing_attrs(data)
        data = pg_crop_number_icon(data)
        data = replace_color(data, color_app)
        self.nr3_large = load_svg_image(data, self._svg_big_number_scale)

        data = pkgutil.get_data(_pkg_name_, f'data/numeric-4{pg_suffix}')
        data = pg_inject_missing_attrs(data)
        data = pg_crop_number_icon(data)
        data = replace_color(data, color_app)
        self.nr4_large = load_svg_image(data, self._svg_big_number_scale)

        data = pkgutil.get_data(_pkg_name_, f'data/numeric-5{pg_suffix}')
        data = pg_inject_missing_attrs(data)
        data = pg_crop_number_icon(data)
        data = replace_color(data, color_app)
        self.nr5_large = load_svg_image(data, self._svg_big_number_scale)

        data = pkgutil.get_data(_pkg_name_, f'data/numeric-6{pg_suffix}')
        data = pg_inject_missing_attrs(data)
        data = pg_crop_number_icon(data)
        data = replace_color(data, color_app)
        self.nr6_large = load_svg_image(data, self._svg_big_number_scale)

        data = pkgutil.get_data(_pkg_name_, f'data/circle-small{pg_suffix}')
        data = pg_inject_missing_attrs(data)
        data = replace_color(data, color_preview)
        self.opt_dot = load_svg_image(data, self.svg_scale)

        data = pkgutil.get_data(_pkg_name_, f'data/dots-hexagon{pg_suffix}')
        data = pg_inject_missing_attrs(data)
        data = replace_color(data, color_preview)
        self.opt_hexagon = load_svg_image(data, self.svg_scale)

        data = pkgutil.get_data(_pkg_name_, f'data/dots-grid{pg_suffix}')
        data = pg_inject_missing_attrs(data)
        data = replace_color(data, color_preview)
        self.opt_square = load_svg_image(data, self.svg_scale)

        data = pkgutil.get_data(_pkg_name_, f'data/dots-triangle{pg_suffix}')
        data = pg_inject_missing_attrs(data)
        data = replace_color(data, color_preview)
        self.opt_triangle = load_svg_image(data, self.svg_scale)

        data = pkgutil.get_data(_pkg_name_, f'data/rotate-orbit{pg_suffix}')
        data = pg_inject_missing_attrs(data)
        data = replace_color(data, color_preview)
        self.fit_3d = load_svg_image(data, self.svg_scale)



import numpy.typing as npt
from matplotlib.figure import Figure
from mpl_toolkits.mplot3d.axes3d import Axes3D

from ..graphs import add_watermark


def draw_3d_locs(fig: Figure,
                 xyz: npt.NDArray[float],
                 set_default_size: bool = True
                 ) -> Figure:
    fig.clear(True)
    if set_default_size:
        fig.set_size_inches(5, 5)
    fig.set_layout_engine('tight')

    ax: Axes3D = fig.add_subplot(projection='3d')
    ax.set_xlabel(r'X [$\mu$m]')
    ax.set_ylabel(r'Y [$\mu$m]')
    ax.set_zlabel(r'Z [$\mu$m]')
    ax.scatter(xyz[:, 0], xyz[:, 1], xyz[:, 2],
               s=1, c=xyz[:, 2], marker='o')

    # ax.view_init(azim=-75, elev=15)  # Doesn't work with tight layout
    ax.set_aspect('equal')
    add_watermark(fig)
    return fig



import time
import tkinter as tk
import traceback as tb
from idlelib.tooltip import Hovertip
from threading import Thread
from tkinter import messagebox, ttk
from typing import Optional

import numpy as np
from matplotlib.figure import Figure

import smlfm
import smlfm.graphs

from .app_model import AppModel, IStage
from .consts import COMMAND, IMAGE, TEXTVARIABLE, VARIABLE, GraphType, StageType
from .figure_window import FigureWindow
from .optics_cfg_dialog import OpticsCfgDialog


# pylint: disable=too-many-ancestors
class OpticsFrame(ttk.Frame, IStage):

    def __init__(self, master, model: AppModel, *args, **kwargs):
        super().__init__(master, *args, **kwargs)

        self._stage_type = StageType.OPTICS
        self._stage_type_next = StageType.FILTER

        self._model = model

        # Controls
        self._ui_frm_lbl = ttk.Label(self)
        self._ui_frm = ttk.LabelFrame(self, labelwidget=self._ui_frm_lbl)

        self._ui_summary = ttk.Frame(self._ui_frm)
        self._ui_settings = ttk.Checkbutton(self._ui_summary, style='Toolbutton')
        self._ui_settings_tip = Hovertip(self._ui_settings, text='')
        self._ui_start = ttk.Checkbutton(self._ui_summary, style='Toolbutton')
        self._ui_start_tip = Hovertip(self._ui_start, text='')
        self._ui_summary_lbl = ttk.Label(self._ui_summary)
        self._ui_preview_mla = ttk.Checkbutton(self._ui_summary, style='Toolbutton')
        self._ui_preview_mla_tip = Hovertip(self._ui_preview_mla, text='')
        self._ui_preview = ttk.Checkbutton(self._ui_summary, style='Toolbutton')
        self._ui_preview_tip = Hovertip(self._ui_preview, text='')

        # Layout
        self._ui_settings.grid(column=0, row=0, sticky=tk.NW)
        self._ui_start.grid(column=1, row=0, sticky=tk.NW)
        self._ui_summary_lbl.grid(column=1, row=0, sticky=tk.EW)
        self._ui_preview_mla.grid(column=2, row=0, sticky=tk.NW)
        self._ui_preview.grid(column=3, row=0, sticky=tk.NW)
        self._ui_summary.columnconfigure(index=1, weight=1)

        self._ui_summary.grid(column=0, row=0, sticky=tk.EW)
        self._ui_frm.columnconfigure(index=0, weight=1)

        self._ui_frm.grid(column=0, row=0, sticky=tk.EW)
        self.columnconfigure(index=0, weight=1)

        # Padding
        self._model.grid_slaves_configure(self._ui_summary, padx=1, pady=1)
        self._model.grid_slaves_configure(self._ui_frm, padx=5, pady=(0, 5))
        self._ui_frm.configure(padding=0)

        # Texts & icons
        self._ui_frm_lbl.configure(text='Micro-Lens Array & Microscope',
                                   image=self._model.icons.opt_stage,
                                   compound=tk.LEFT)
        self._ui_settings[IMAGE] = self._model.icons.settings
        self._ui_settings_tip.text = 'Edit properties'
        self._ui_start[IMAGE] = self._model.icons.start
        self._ui_start_tip.text = 'Proceed with configuration'
        self._ui_preview_mla[IMAGE] = self._model.icons.opt_dot
        self._ui_preview_mla_tip.text = 'Preview lens array'
        self._ui_preview[IMAGE] = self._model.icons.preview
        self._ui_preview_tip.text = 'Preview data mapping'

        # Hooks
        self._ui_settings[COMMAND] = self._on_settings
        self._ui_start[COMMAND] = self._on_start
        self._ui_summary.bind(
            '<Configure>', lambda _e: self._ui_summary_lbl.configure(
                wraplength=self._ui_summary_lbl.winfo_width()))
        self._ui_preview_mla[COMMAND] = self._on_preview_mla
        self._ui_preview[COMMAND] = self._on_preview

        # Data
        self._var_settings = tk.IntVar()
        self._ui_settings[VARIABLE] = self._var_settings
        self._var_start = tk.IntVar()
        self._ui_start[VARIABLE] = self._var_start
        self._var_summary = tk.StringVar()
        self._ui_summary_lbl[TEXTVARIABLE] = self._var_summary
        self._var_preview_mla = tk.IntVar()
        self._ui_preview_mla[VARIABLE] = self._var_preview_mla
        self._var_preview = tk.IntVar()
        self._ui_preview[VARIABLE] = self._var_preview

        self._update_thread: Optional[Thread] = None
        self._update_thread_err: Optional[str] = None

        self._settings_dlg: Optional[OpticsCfgDialog] = None
        self._settings_apply: bool = False
        self._update_from_settings: bool = False

        # =0 ~ disabled, >0 ~ remaining flashes, None ~ infinite
        self._flash_counter = 0
        self._flash_id = None

        self.stage_ui_init()

    def stage_type(self):
        return self._stage_type

    def stage_type_next(self):
        return self._stage_type_next

    def stage_is_active(self) -> bool:
        csv = self._model.csv
        return csv is not None and csv.data is not None and csv.data.shape[0] > 0

    def stage_invalidate(self):
        self._model.lfm = None
        self._model.mla = None
        self._model.lfl = None

        if not self._update_from_settings:
            for gt in [GraphType.MLA, GraphType.MAPPED]:
                self._model.destroy_graph(gt)

        self._model.stage_enabled(self._stage_type, False)

        self.stage_ui_init()
        self._model.stage_invalidate(self._stage_type_next)

    def stage_start_update(self):
        self.stage_invalidate()
        self._model.stages_ui_updating(True)
        self.winfo_toplevel().configure(cursor='watch')

        def _update_thread_fn():
            try:
                self._update_task()
            except BaseException as ex:
                self._update_thread_err = str(ex)
                tb.print_exception(None, ex, ex.__traceback__)

            def _update_done():
                self._update_thread = None
                self.winfo_toplevel().configure(cursor='')

                self._model.stages_ui_updating(False)

                if self._update_thread_err is not None:
                    err = self._update_thread_err
                    self._update_thread_err = None

                    messagebox.showerror(
                        title='Configuration Error',
                        message=f'The optics cannot be configured:\n{err}')

                    self._settings_apply = False
                    self._update_from_settings = False

                    self.stage_invalidate()
                    return

                if self._model.stage_is_active(self._stage_type_next):

                    self._show_previews(force_update=True)

                    self._model.stage_ui_init(self._stage_type_next)
                    # Next phase updates automatically or the user continues manually
                    if self._update_from_settings:
                        self._model.stage_ui_flash(self._stage_type_next)
                    else:
                        if self._model.cfg.confirm_mla_alignment:
                            self._model.stage_ui_flash(self._stage_type_next)
                            if self._model.cfg.show_mla_alignment_graph:
                                messagebox.showinfo(
                                    title='Confirmation',
                                    message='Check on the figure that the lenses'
                                            ' are well aligned with the data. If'
                                            ' the alignment is incorrect, adjust'
                                            ' Micro-Lens Array parameters.\n'
                                            '\n'
                                            'Otherwise continue with flashing button.')
                        else:
                            self._model.stage_start_update(self._stage_type_next)

                self._settings_apply = False
                self._update_from_settings = False

            self._model.invoke_on_gui_thread_async(_update_done)

        self._update_thread = Thread(target=_update_thread_fn, daemon=True)
        self._update_thread.start()

    def stage_ui_init(self):
        self._ui_update_done()

    def stage_ui_flash(self):
        self._flash_start()

    def stage_ui_updating(self, updating: bool):
        if updating:
            self._ui_update_start()
        else:
            self._ui_update_done()

    def _ui_update_start(self):
        self._ui_settings.configure(state=tk.DISABLED)
        if self._update_from_settings:
            settings_dlg = self._settings_dlg
            if settings_dlg is not None:
                settings_dlg.enable(False)
        self._ui_start.configure(state=tk.DISABLED)
        self._ui_preview_mla.configure(state=tk.DISABLED)
        self._ui_preview.configure(state=tk.DISABLED)

    def _ui_update_done(self):
        self._ui_update_start()

        if self.stage_is_active():
            self._model.stage_enabled(self._stage_type, True)

            self._ui_settings.configure(state=tk.NORMAL)
            if self._update_from_settings:
                settings_dlg = self._settings_dlg
                if settings_dlg is not None:
                    settings_dlg.enable(True)
            self._ui_start.configure(state=tk.NORMAL)

            if self._model.stage_is_active(self._stage_type_next):
                self._ui_preview_mla.configure(state=tk.NORMAL)

                if (self._model.mla.lattice_type
                        == smlfm.MicroLensArray.LatticeType.HEXAGONAL):
                    self._ui_preview_mla[IMAGE] = self._model.icons.opt_hexagon
                elif (self._model.mla.lattice_type
                      == smlfm.MicroLensArray.LatticeType.SQUARE):
                    self._ui_preview_mla[IMAGE] = self._model.icons.opt_square
                # elif (self._model.mla.lattice_type
                #       == smlfm.MicroLensArray.LatticeType.TRIANGULAR):
                #     self._ui_preview_mla[IMAGE] = self._model.icons.opt_triangle
                else:
                    self._ui_preview_mla[IMAGE] = self._model.icons.opt_dot

                if self._model.lfl is not None:
                    self._ui_preview.configure(state=tk.NORMAL)
                    lenses = np.unique(self._model.lfl.locs_2d[:, 12]).shape[0]
                    self._var_summary.set(f'Localisations mapped to {lenses} lenses')
                else:
                    self._var_summary.set('No localisations mapped to lenses')
            else:
                self._var_summary.set('No optics configured')
        else:
            self._var_summary.set('No optics configured yet')

    def _flash_start(self):
        self._flash_stop()

        def _toggle_on():
            if self._flash_counter is None or self._flash_counter > 0:
                self._var_start.set(1)
                self._flash_id = self.after(500, _toggle_off)

        def _toggle_off():
            self._var_start.set(0)
            if self._flash_counter is None or self._flash_counter > 0:
                if self._flash_counter is not None:
                    self._flash_counter -= 1
                self._flash_id = self.after(500, _toggle_on)

        self._flash_counter = None
        _toggle_on()

    def _flash_stop(self):
        self._flash_counter = 0
        if self._flash_id is not None:
            self.after_cancel(self._flash_id)
        self._var_start.set(0)

    def _on_settings(self):
        cfg_dump_old = self._model.cfg.to_json()

        def _process_cb(apply: bool):
            self._settings_apply = apply
            self._update_from_settings = True
            nonlocal cfg_dump_old
            cfg_dump_new = self._model.cfg.to_json()
            if (cfg_dump_old != cfg_dump_new
                    or not self._model.stage_is_active(self._stage_type_next)):
                cfg_dump_old = cfg_dump_new
                self.stage_start_update()
            else:
                self._show_previews()
                self._settings_apply = False
                self._update_from_settings = False

        self._settings_dlg = OpticsCfgDialog(self, self._model, process_cb=_process_cb)
        self._settings_dlg.show_modal()
        self._settings_dlg = None
        if not self._var_preview_mla.get() and not self._var_preview.get():
            self._settings_apply = False
            self._update_from_settings = False
        self._var_settings.set(0)

    def _on_start(self):
        self._var_start.set(0)

        self._flash_stop()
        self.stage_start_update()

    def _show_previews(self, force_update: bool = False):
        if not self._model.stage_is_active(self._stage_type_next):
            return

        show = self._model.cfg.show_graphs

        force_show_mla = ((show and self._model.cfg.show_all_debug_graphs)
                          or self._settings_apply)
        force_update_mla = (force_update
                            and self._model.graph_exists(GraphType.MLA))
        if force_show_mla:
            self._var_preview_mla.set(1)
        if force_show_mla or force_update_mla:
            self._on_preview_mla(force_update_mla)

        force_show_map = ((show and self._model.cfg.show_mla_alignment_graph)
                          or self._settings_apply)
        force_update_map = (force_update
                            and self._model.graph_exists(GraphType.MAPPED))
        if force_show_map:
            self._var_preview.set(1)
        if force_show_map or force_update_map:
            self._on_preview(force_update_map)

    def _on_preview_mla(self, force_update: bool = False):

        def draw(f: Figure, set_size: bool = True) -> Figure:
            return smlfm.graphs.draw_mla(
                f,
                lens_centres=self._model.mla.lens_centres,
                mla_centre=self._model.mla.centre,
                bfp_radius=self._model.lfm.bfp_radius / self._model.mla.lens_pitch,
                set_default_size=set_size)

        wnd = self._model.graphs[GraphType.MLA]
        if wnd is None:
            fig = draw(Figure())
            wnd = FigureWindow(fig, master=self,
                               title='Micro-lens array centres')
            wnd.bind('<Map>', func=lambda _evt: self._var_preview_mla.set(1))
            wnd.bind('<Unmap>', func=lambda _evt: self._var_preview_mla.set(0))
            self._model.graphs[GraphType.MLA] = wnd
        else:
            if force_update:
                draw(wnd.figure, set_size=False)
                wnd.refresh()

        if self._var_preview_mla.get():
            wnd.deiconify()
        else:
            wnd.withdraw()

    def _on_preview(self, force_update: bool = False):

        def draw(f: Figure, set_size: bool = True) -> Figure:
            return smlfm.graphs.draw_locs(
                f,
                xy=self._model.lfl.locs_2d[:, 3:5],
                lens_idx=self._model.lfl.locs_2d[:, 12],
                lens_centres=((self._model.mla.lens_centres - self._model.mla.centre)
                              * self._model.lfm.mla_to_xy_scale),
                # U,V values are around MLA centre but that offset is not included
                # mla_centre=self._model.mla.centre,
                mla_centre=np.array([0.0, 0.0]),
                bfp_radius=((self._model.lfm.bfp_radius / self._model.mla.lens_pitch)
                            * self._model.lfm.mla_to_xy_scale),
                set_default_size=set_size)

        wnd = self._model.graphs[GraphType.MAPPED]
        if wnd is None:
            fig = draw(Figure())
            wnd = FigureWindow(fig, master=self,
                               title='Localisations with lens centers')
            wnd.bind('<Map>', func=lambda _evt: self._var_preview.set(1))
            wnd.bind('<Unmap>', func=lambda _evt: self._var_preview.set(0))
            self._model.graphs[GraphType.MAPPED] = wnd
        else:
            if force_update:
                draw(wnd.figure, set_size=False)
                wnd.refresh()

        if self._var_preview.get():
            wnd.deiconify()
        else:
            wnd.withdraw()

    # Executed on thread
    def _update_task(self):
        self._model.invoke_on_gui_thread_async(
            self._var_summary.set, 'Configuring optics...')

        self._model.lfm = smlfm.FourierMicroscope(
            self._model.cfg.num_aperture,
            self._model.cfg.mla_lens_pitch,
            self._model.cfg.focal_length_mla,
            self._model.cfg.focal_length_obj_lens,
            self._model.cfg.focal_length_tube_lens,
            self._model.cfg.focal_length_fourier_lens,
            self._model.cfg.pixel_size_camera,
            self._model.cfg.ref_idx_immersion,
            self._model.cfg.ref_idx_medium)

        self._model.mla = smlfm.MicroLensArray(
            self._model.cfg.mla_type,
            self._model.cfg.focal_length_mla,
            self._model.cfg.mla_lens_pitch,
            self._model.cfg.mla_optic_size,
            np.array(self._model.cfg.mla_centre))

        self._model.mla.rotate_lattice(np.deg2rad(self._model.cfg.mla_rotation))
        self._model.mla.offset_lattice(
            np.array(self._model.cfg.mla_offset) / self._model.lfm.mla_to_xy_scale)  # XY -> MLA

        self._model.invoke_on_gui_thread_async(
            self._var_summary.set, 'Mapping to lenses...')
        tic = time.time()

        self._model.csv.scale_peakfit_data(self._model.lfm.pixel_size_sample)
        locs_2d = self._model.csv.data.copy()

        # Center X and Y to their means
        # TODO: Why center the data to means? Ensure how is this related to MLA centre.
        locs_2d[:, 1] -= locs_2d[:, 1].mean()
        locs_2d[:, 2] -= locs_2d[:, 2].mean()

        self._model.lfl = smlfm.Localisations(locs_2d)
        self._model.lfl.assign_to_lenses(self._model.mla, self._model.lfm)

        if self._model.cfg.log_timing:
            print(f'Mapping points to lenses took'
                  f' {1e3 * (time.time() - tic):.3f} ms')



from .add_watermark import add_watermark
from .draw_3d_locs import draw_3d_locs
from .draw_histogram import draw_histogram
from .draw_locs import draw_locs
from .draw_locs_csv import draw_locs_csv
from .draw_locs_normalised import draw_locs_normalised
from .draw_mla import draw_mla
from .draw_occurrences import draw_occurrences
from .reconstruct_results import reconstruct_results

__all__ = [
    "add_watermark",
    "draw_3d_locs",
    "draw_histogram",
    "draw_locs",
    "draw_locs_csv",
    "draw_locs_normalised",
    "draw_mla",
    "draw_occurrences",
    "reconstruct_results",
]



import tkinter as tk
from idlelib.tooltip import Hovertip
from tkinter import ttk
from typing import Callable, Optional

import smlfm
from .app_model import AppModel
from .cfg_dialog import CfgDialog
from .consts import READONLY


# pylint: disable=too-many-instance-attributes
class OpticsCfgDialog(CfgDialog):

    def __init__(self, parent, model: AppModel, title: Optional[str] = None,
                 process_cb: Optional[Callable[[bool], None]] = None):
        if title is None:
            title = 'Optics Settings'

        self._ui_mla_type: Optional[ttk.Combobox] = None
        self._var_mla_type = tk.StringVar(value=model.cfg.mla_type.name)

        self._ui_mla_lens_pitch: Optional[ttk.Entry] = None
        self._var_mla_lens_pitch = tk.StringVar(value=str(model.cfg.mla_lens_pitch))
        self._ui_mla_lens_pitch_tip: Optional[Hovertip] = None

        self._ui_mla_optic_size: Optional[ttk.Entry] = None
        self._var_mla_optic_size = tk.StringVar(value=str(model.cfg.mla_optic_size))
        self._ui_mla_optic_size_tip: Optional[Hovertip] = None

        self._ui_mla_centre_x: Optional[ttk.Entry] = None
        self._var_mla_centre_x = tk.StringVar(value=str(model.cfg.mla_centre[0]))
        self._ui_mla_centre_x_tip: Optional[Hovertip] = None
        self._ui_mla_centre_y: Optional[ttk.Entry] = None
        self._var_mla_centre_y = tk.StringVar(value=str(model.cfg.mla_centre[1]))
        self._ui_mla_centre_y_tip: Optional[Hovertip] = None

        self._ui_mla_rotation: Optional[ttk.Entry] = None
        self._var_mla_rotation = tk.StringVar(value=str(model.cfg.mla_rotation))
        self._ui_mla_rotation_tip: Optional[Hovertip] = None

        self._ui_mla_offset_x: Optional[ttk.Entry] = None
        self._var_mla_offset_x = tk.StringVar(value=str(model.cfg.mla_offset[0]))
        self._ui_mla_offset_x_tip: Optional[Hovertip] = None
        self._ui_mla_offset_y: Optional[ttk.Entry] = None
        self._var_mla_offset_y = tk.StringVar(value=str(model.cfg.mla_offset[1]))
        self._ui_mla_offset_y_tip: Optional[Hovertip] = None

        self._ui_focal_length_mla: Optional[ttk.Entry] = None
        self._var_focal_length_mla = tk.StringVar(
            value=str(model.cfg.focal_length_mla))
        self._ui_focal_length_mla_tip: Optional[Hovertip] = None

        self._ui_focal_length_obj_lens: Optional[ttk.Entry] = None
        self._var_focal_length_obj_lens = tk.StringVar(
            value=str(model.cfg.focal_length_obj_lens))
        self._ui_focal_length_obj_lens_tip: Optional[Hovertip] = None

        self._ui_focal_length_fourier_lens: Optional[ttk.Entry] = None
        self._var_focal_length_fourier_lens = tk.StringVar(
            value=str(model.cfg.focal_length_fourier_lens))
        self._ui_focal_length_fourier_lens_tip: Optional[Hovertip] = None

        self._ui_focal_length_tube_lens: Optional[ttk.Entry] = None
        self._var_focal_length_tube_lens = tk.StringVar(
            value=str(model.cfg.focal_length_tube_lens))
        self._ui_focal_length_tube_lens_tip: Optional[Hovertip] = None

        self._ui_num_aperture: Optional[ttk.Entry] = None
        self._var_num_aperture = tk.StringVar(value=str(model.cfg.num_aperture))
        self._ui_num_aperture_tip: Optional[Hovertip] = None

        self._ui_ref_idx_immersion: Optional[ttk.Entry] = None
        self._var_ref_idx_immersion = tk.StringVar(
            value=str(model.cfg.ref_idx_immersion))
        self._ui_ref_idx_immersion_tip: Optional[Hovertip] = None

        self._ui_ref_idx_medium: Optional[ttk.Entry] = None
        self._var_ref_idx_medium = tk.StringVar(value=str(model.cfg.ref_idx_medium))
        self._ui_ref_idx_medium_tip: Optional[Hovertip] = None

        self._ui_pixel_size_camera: Optional[ttk.Entry] = None
        self._var_pixel_size_camera = tk.StringVar(
            value=str(model.cfg.pixel_size_camera))
        self._ui_pixel_size_camera_tip: Optional[Hovertip] = None

        super().__init__(parent, model, title, process_cb=process_cb)

    # noinspection PyProtectedMember
    # pylint: disable=protected-access,too-many-statements
    def body(self, master) -> tk.BaseWidget:
        ui_tab = ttk.Frame(master)
        row = 0

        ui_mla_type_lbl = ttk.Label(ui_tab, text='MLA lattice type:', anchor=tk.E)
        Hovertip(ui_mla_type_lbl, text=self.model.cfg._mla_type_doc)
        self._ui_mla_type = ttk.Combobox(
            ui_tab, state=READONLY,
            values=[t.name for t in smlfm.MicroLensArray.LatticeType],
            textvariable=self._var_mla_type)
        Hovertip(self._ui_mla_type, text=self.model.cfg._mla_type_doc)
        self.ui_widgets.append(self._ui_mla_type)
        #
        ui_mla_type_lbl.grid(column=0, row=row, sticky=tk.W)
        self._ui_mla_type.grid(column=1, row=row, sticky=tk.EW, columnspan=2)
        row += 1

        ui_mla_lens_pitch_lbl = ttk.Label(ui_tab, text='MLA lens pitch:', anchor=tk.E)
        Hovertip(ui_mla_lens_pitch_lbl, text=self.model.cfg._mla_lens_pitch_doc)
        self._ui_mla_lens_pitch = ttk.Entry(
            ui_tab, textvariable=self._var_mla_lens_pitch)
        self._ui_mla_lens_pitch_tip = Hovertip(
            self._ui_mla_lens_pitch, text=self.model.cfg._mla_lens_pitch_doc)
        self.ui_widgets.append(self._ui_mla_lens_pitch)
        ui_mla_lens_pitch_unit = ttk.Label(ui_tab, text='\u00B5m')  # microns
        #
        ui_mla_lens_pitch_lbl.grid(column=0, row=row, sticky=tk.W)
        self._ui_mla_lens_pitch.grid(column=1, row=row, sticky=tk.EW, columnspan=2)
        ui_mla_lens_pitch_unit.grid(column=3, row=row, sticky=tk.W)
        row += 1

        ui_mla_optic_size_lbl = ttk.Label(ui_tab, text='MLA optic size:', anchor=tk.E)
        Hovertip(ui_mla_optic_size_lbl, text=self.model.cfg._mla_optic_size_doc)
        self._ui_mla_optic_size = ttk.Entry(
            ui_tab, textvariable=self._var_mla_optic_size)
        self._ui_mla_optic_size_tip = Hovertip(
            self._ui_mla_optic_size, text=self.model.cfg._mla_optic_size_doc)
        self.ui_widgets.append(self._ui_mla_optic_size)
        ui_mla_optic_size_unit = ttk.Label(ui_tab, text='\u00B5m')  # microns
        #
        ui_mla_optic_size_lbl.grid(column=0, row=row, sticky=tk.W)
        self._ui_mla_optic_size.grid(column=1, row=row, sticky=tk.EW, columnspan=2)
        ui_mla_optic_size_unit.grid(column=3, row=row, sticky=tk.W)
        row += 1

        ui_mla_centre_lbl = ttk.Label(ui_tab, text='MLA centre:', anchor=tk.E)
        Hovertip(ui_mla_centre_lbl, text=self.model.cfg._mla_centre_doc)
        self._ui_mla_centre_x = ttk.Entry(
            ui_tab, textvariable=self._var_mla_centre_x)
        self._ui_mla_centre_x_tip = Hovertip(
            self._ui_mla_centre_x, text=self.model.cfg._mla_centre_doc)
        self.ui_widgets.append(self._ui_mla_centre_x)
        self._ui_mla_centre_y = ttk.Entry(
            ui_tab, textvariable=self._var_mla_centre_y)
        self._ui_mla_centre_y_tip = Hovertip(
            self._ui_mla_centre_y, text=self.model.cfg._mla_centre_doc)
        self.ui_widgets.append(self._ui_mla_centre_y)
        ui_mla_centre_unit = ttk.Label(ui_tab, text='[lsu]')
        Hovertip(ui_mla_centre_unit, text='lattice spacing units')
        #
        ui_mla_centre_lbl.grid(column=0, row=row, sticky=tk.W)
        self._ui_mla_centre_x.grid(column=1, row=row, sticky=tk.EW)
        self._ui_mla_centre_y.grid(column=2, row=row, sticky=tk.EW)
        ui_mla_centre_unit.grid(column=3, row=row, sticky=tk.W)
        row += 1

        ui_mla_rotation_lbl = ttk.Label(ui_tab, text='MLA rotation:', anchor=tk.E)
        Hovertip(ui_mla_rotation_lbl, text=self.model.cfg._mla_rotation_doc)
        self._ui_mla_rotation = ttk.Entry(
            ui_tab, textvariable=self._var_mla_rotation)
        self._ui_mla_rotation_tip = Hovertip(
            self._ui_mla_rotation, text=self.model.cfg._mla_rotation_doc)
        self.ui_widgets.append(self._ui_mla_rotation)
        ui_mla_rotation_unit = ttk.Label(ui_tab, text='deg')
        #
        ui_mla_rotation_lbl.grid(column=0, row=row, sticky=tk.W)
        self._ui_mla_rotation.grid(column=1, row=row, sticky=tk.EW, columnspan=2)
        ui_mla_rotation_unit.grid(column=3, row=row, sticky=tk.W)
        row += 1

        ui_mla_offset_lbl = ttk.Label(ui_tab, text='MLA offset:', anchor=tk.E)
        Hovertip(ui_mla_offset_lbl, text=self.model.cfg._mla_offset_doc)
        self._ui_mla_offset_x = ttk.Entry(
            ui_tab, textvariable=self._var_mla_offset_x)
        self._ui_mla_offset_x_tip = Hovertip(
            self._ui_mla_offset_x, text=self.model.cfg._mla_offset_doc)
        self.ui_widgets.append(self._ui_mla_offset_x)
        self._ui_mla_offset_y = ttk.Entry(
            ui_tab, textvariable=self._var_mla_offset_y)
        self._ui_mla_offset_y_tip = Hovertip(
            self._ui_mla_offset_y, text=self.model.cfg._mla_offset_doc)
        self.ui_widgets.append(self._ui_mla_offset_y)
        ui_mla_offset_unit = ttk.Label(ui_tab, text='\u00B5m')  # microns
        #
        ui_mla_offset_lbl.grid(column=0, row=row, sticky=tk.W)
        self._ui_mla_offset_x.grid(column=1, row=row, sticky=tk.EW)
        self._ui_mla_offset_y.grid(column=2, row=row, sticky=tk.EW)
        ui_mla_offset_unit.grid(column=3, row=row, sticky=tk.W)
        row += 1

        ui_focal_length_mla_lbl = ttk.Label(
            ui_tab, text='MLA focal length:', anchor=tk.E)
        Hovertip(ui_focal_length_mla_lbl, text=self.model.cfg._focal_length_mla_doc)
        self._ui_focal_length_mla = ttk.Entry(
            ui_tab, textvariable=self._var_focal_length_mla)
        self._ui_focal_length_mla_tip = Hovertip(
            self._ui_focal_length_mla, text=self.model.cfg._focal_length_mla_doc)
        self.ui_widgets.append(self._ui_focal_length_mla)
        ui_focal_length_mla_unit = ttk.Label(ui_tab, text='mm')
        #
        ui_focal_length_mla_lbl.grid(column=0, row=row, sticky=tk.W)
        self._ui_focal_length_mla.grid(column=1, row=row, sticky=tk.EW, columnspan=2)
        ui_focal_length_mla_unit.grid(column=3, row=row, sticky=tk.W)
        row += 1

        ui_focal_length_obj_lens_lbl = ttk.Label(
            ui_tab, text='Objective lens focal length:', anchor=tk.E)
        Hovertip(ui_focal_length_obj_lens_lbl,
                 text=self.model.cfg._focal_length_obj_lens_doc)
        self._ui_focal_length_obj_lens = ttk.Entry(
            ui_tab, textvariable=self._var_focal_length_obj_lens)
        self._ui_focal_length_obj_lens_tip = Hovertip(
            self._ui_focal_length_obj_lens,
            text=self.model.cfg._focal_length_obj_lens_doc)
        self.ui_widgets.append(self._ui_focal_length_obj_lens)
        ui_focal_length_obj_lens_unit = ttk.Label(ui_tab, text='mm')
        #
        ui_focal_length_obj_lens_lbl.grid(column=0, row=row, sticky=tk.W)
        self._ui_focal_length_obj_lens.grid(column=1, row=row, sticky=tk.EW, columnspan=2)
        ui_focal_length_obj_lens_unit.grid(column=3, row=row, sticky=tk.W)
        row += 1

        ui_focal_length_fourier_lens_lbl = ttk.Label(
            ui_tab, text='Fourier lens focal length:', anchor=tk.E)
        Hovertip(ui_focal_length_fourier_lens_lbl,
                 text=self.model.cfg._focal_length_fourier_lens_doc)
        self._ui_focal_length_fourier_lens = ttk.Entry(
            ui_tab, textvariable=self._var_focal_length_fourier_lens)
        self._ui_focal_length_fourier_lens_tip = Hovertip(
            self._ui_focal_length_fourier_lens,
            text=self.model.cfg._focal_length_fourier_lens_doc)
        self.ui_widgets.append(self._ui_focal_length_fourier_lens)
        ui_focal_length_fourier_lens_unit = ttk.Label(ui_tab, text='mm')
        #
        ui_focal_length_fourier_lens_lbl.grid(column=0, row=row, sticky=tk.W)
        self._ui_focal_length_fourier_lens.grid(column=1, row=row, sticky=tk.EW, columnspan=2)
        ui_focal_length_fourier_lens_unit.grid(column=3, row=row, sticky=tk.W)
        row += 1

        ui_focal_length_tube_lens_lbl = ttk.Label(
            ui_tab, text='Tube lens focal length:', anchor=tk.E)
        Hovertip(ui_focal_length_tube_lens_lbl,
                 text=self.model.cfg._focal_length_tube_lens_doc)
        self._ui_focal_length_tube_lens = ttk.Entry(
            ui_tab, textvariable=self._var_focal_length_tube_lens)
        self._ui_focal_length_tube_lens_tip = Hovertip(
            self._ui_focal_length_tube_lens,
            text=self.model.cfg._focal_length_tube_lens_doc)
        self.ui_widgets.append(self._ui_focal_length_tube_lens)
        ui_focal_length_tube_lens_unit = ttk.Label(ui_tab, text='mm')
        #
        ui_focal_length_tube_lens_lbl.grid(column=0, row=row, sticky=tk.W)
        self._ui_focal_length_tube_lens.grid(column=1, row=row, sticky=tk.EW, columnspan=2)
        ui_focal_length_tube_lens_unit.grid(column=3, row=row, sticky=tk.W)
        row += 1

        ui_num_aperture_lbl = ttk.Label(
            ui_tab, text='Objective num. aperture:', anchor=tk.E)
        Hovertip(ui_num_aperture_lbl, text=self.model.cfg._num_aperture_doc)
        self._ui_num_aperture = ttk.Entry(
            ui_tab, textvariable=self._var_num_aperture)
        self._ui_num_aperture_tip = Hovertip(
            self._ui_num_aperture, text=self.model.cfg._num_aperture_doc)
        self.ui_widgets.append(self._ui_num_aperture)
        #
        ui_num_aperture_lbl.grid(column=0, row=row, sticky=tk.W)
        self._ui_num_aperture.grid(column=1, row=row, sticky=tk.EW, columnspan=2)
        row += 1

        ui_ref_idx_immersion_lbl = ttk.Label(
            ui_tab, text='Immersion refractive index:', anchor=tk.E)
        Hovertip(ui_ref_idx_immersion_lbl, text=self.model.cfg._ref_idx_immersion_doc)
        self._ui_ref_idx_immersion = ttk.Entry(
            ui_tab, textvariable=self._var_ref_idx_immersion)
        self._ui_ref_idx_immersion_tip = Hovertip(
            self._ui_ref_idx_immersion, text=self.model.cfg._ref_idx_immersion_doc)
        self.ui_widgets.append(self._ui_ref_idx_immersion)
        #
        ui_ref_idx_immersion_lbl.grid(column=0, row=row, sticky=tk.W)
        self._ui_ref_idx_immersion.grid(column=1, row=row, sticky=tk.EW, columnspan=2)
        row += 1

        ui_ref_idx_medium_lbl = ttk.Label(
            ui_tab, text='Medium refractive index:', anchor=tk.E)
        Hovertip(ui_ref_idx_medium_lbl, text=self.model.cfg._ref_idx_medium_doc)
        self._ui_ref_idx_medium = ttk.Entry(
            ui_tab, textvariable=self._var_ref_idx_medium)
        self._ui_ref_idx_medium_tip = Hovertip(
            self._ui_ref_idx_medium, text=self.model.cfg._ref_idx_medium_doc)
        self.ui_widgets.append(self._ui_ref_idx_medium)
        #
        ui_ref_idx_medium_lbl.grid(column=0, row=row, sticky=tk.W)
        self._ui_ref_idx_medium.grid(column=1, row=row, sticky=tk.EW, columnspan=2)
        row += 1

        ui_pixel_size_camera_lbl = ttk.Label(
            ui_tab, text='Camera pixel size:', anchor=tk.E)
        Hovertip(ui_pixel_size_camera_lbl, text=self.model.cfg._pixel_size_camera_doc)
        self._ui_pixel_size_camera = ttk.Entry(
            ui_tab, textvariable=self._var_pixel_size_camera)
        self._ui_pixel_size_camera_tip = Hovertip(
            self._ui_pixel_size_camera, text=self.model.cfg._pixel_size_camera_doc)
        self.ui_widgets.append(self._ui_pixel_size_camera)
        ui_pixel_size_camera_unit = ttk.Label(ui_tab, text='\u00B5m')  # microns
        #
        ui_pixel_size_camera_lbl.grid(column=0, row=row, sticky=tk.W)
        self._ui_pixel_size_camera.grid(column=1, row=row, sticky=tk.EW, columnspan=2)
        ui_pixel_size_camera_unit.grid(column=3, row=row, sticky=tk.W)
        row += 1

        # Final padding and layout
        ui_tab.columnconfigure(index=1, weight=1)
        ui_tab.columnconfigure(index=2, weight=1)
        self.model.grid_slaves_configure(ui_tab, padx=1, pady=1)

        ui_tab.pack(anchor=tk.NW, fill=tk.X, expand=True, padx=5, pady=5)

        return self._ui_mla_type  # Control that gets initial focus

    # pylint: disable=too-many-return-statements,too-many-branches
    def validate(self) -> bool:
        if not self.is_float(self._var_mla_lens_pitch.get()):
            self.initial_focus = self._ui_mla_lens_pitch
            self._ui_mla_lens_pitch_tip.showtip()
            return False
        if not self.is_float(self._var_mla_optic_size.get()):
            self.initial_focus = self._ui_mla_optic_size
            self._ui_mla_optic_size_tip.showtip()
            return False
        if not self.is_float(self._var_mla_centre_x.get()):
            self.initial_focus = self._ui_mla_centre_x
            self._ui_mla_centre_x_tip.showtip()
            return False
        if not self.is_float(self._var_mla_centre_y.get()):
            self.initial_focus = self._ui_mla_centre_y
            self._ui_mla_centre_y_tip.showtip()
            return False
        if not self.is_float(self._var_mla_rotation.get()):
            self.initial_focus = self._ui_mla_rotation
            self._ui_mla_rotation_tip.showtip()
            return False
        if not self.is_float(self._var_mla_offset_x.get()):
            self.initial_focus = self._ui_mla_offset_x
            self._ui_mla_offset_x_tip.showtip()
            return False
        if not self.is_float(self._var_mla_offset_y.get()):
            self.initial_focus = self._ui_mla_offset_y
            self._ui_mla_offset_y_tip.showtip()
            return False
        if not self.is_float(self._var_focal_length_mla.get()):
            self.initial_focus = self._ui_focal_length_mla
            self._ui_focal_length_mla_tip.showtip()
            return False
        if not self.is_float(self._var_focal_length_obj_lens.get()):
            self.initial_focus = self._ui_focal_length_obj_lens
            self._ui_focal_length_obj_lens_tip.showtip()
            return False
        if not self.is_float(self._var_focal_length_fourier_lens.get()):
            self.initial_focus = self._ui_focal_length_fourier_lens
            self._ui_focal_length_fourier_lens_tip.showtip()
            return False
        if not self.is_float(self._var_focal_length_tube_lens.get()):
            self.initial_focus = self._ui_focal_length_tube_lens
            self._ui_focal_length_tube_lens_tip.showtip()
            return False
        if not self.is_float(self._var_num_aperture.get()):
            self.initial_focus = self._ui_num_aperture
            self._ui_num_aperture_tip.showtip()
            return False
        if not self.is_float(self._var_ref_idx_immersion.get()):
            self.initial_focus = self._ui_ref_idx_immersion
            self._ui_ref_idx_immersion_tip.showtip()
            return False
        if not self.is_float(self._var_ref_idx_medium.get()):
            self.initial_focus = self._ui_ref_idx_medium
            self._ui_ref_idx_medium_tip.showtip()
            return False
        if not self.is_float(self._var_pixel_size_camera.get()):
            self.initial_focus = self._ui_pixel_size_camera
            self._ui_pixel_size_camera_tip.showtip()
            return False

        return super().validate()

    def process(self) -> None:
        self.model.cfg.mla_type = smlfm.MicroLensArray.LatticeType[
            self._var_mla_type.get()]
        self.model.cfg.mla_lens_pitch = float(self._var_mla_lens_pitch.get())
        self.model.cfg.mla_optic_size = float(self._var_mla_optic_size.get())
        self.model.cfg.mla_centre = (  # Tuple
            float(self._var_mla_centre_x.get()),
            float(self._var_mla_centre_y.get()))
        self.model.cfg.mla_rotation = float(self._var_mla_rotation.get())
        self.model.cfg.mla_offset = (  # Tuple
            float(self._var_mla_offset_x.get()),
            float(self._var_mla_offset_y.get()))
        self.model.cfg.focal_length_mla = float(self._var_focal_length_mla.get())
        self.model.cfg.focal_length_obj_lens = (
            float(self._var_focal_length_obj_lens.get()))
        self.model.cfg.focal_length_fourier_lens = (
            float(self._var_focal_length_fourier_lens.get()))
        self.model.cfg.focal_length_tube_lens = (
            float(self._var_focal_length_tube_lens.get()))
        self.model.cfg.num_aperture = float(self._var_num_aperture.get())
        self.model.cfg.ref_idx_immersion = float(self._var_ref_idx_immersion.get())
        self.model.cfg.ref_idx_medium = float(self._var_ref_idx_medium.get())
        self.model.cfg.pixel_size_camera = float(self._var_pixel_size_camera.get())

        super().process()



#!/usr/bin/env python3

import dataclasses
import pkgutil
import sys
import time
from datetime import datetime
from pathlib import Path

import numpy as np
from matplotlib import pyplot as plt

import smlfm
import smlfm.graphs


# pylint: disable=too-many-branches,too-many-statements
def app():
    tic_total = time.time()
    user_interaction_time = 0
    timestamp = datetime.now()

    # 0. Handle CLI options

    cfg_file = None
    img_stack = None
    csv_file = None
    for arg in sys.argv[1:]:
        a = arg.casefold()
        if a.endswith('.json'):
            cfg_file = Path(arg)
        elif a.endswith('.tiff') or a.endswith('.tif'):
            img_stack = Path(arg)
        elif a.endswith('.csv') or a.endswith('.xls'):
            csv_file = Path(arg)
        else:
            print(f'WARNING: Ignoring unrecognized option "{arg}"')

    # 1. Prepare configuration

    if cfg_file is not None:
        with open(cfg_file, 'rt', encoding='utf-8') as file:
            cfg_dump = file.read()
    else:
        cfg_dump = pkgutil.get_data(
            smlfm.__name__, 'data/default-config.json').decode()
        print('WARNING: Loaded default configuration')

    cfg = smlfm.Config.from_json(cfg_dump)

    # CLI option takes precedence
    if img_stack is not None:
        cfg.img_stack = img_stack
    if csv_file is not None:
        cfg.csv_file = csv_file

    if cfg.img_stack is not None:
        if not cfg.img_stack.exists():
            print(f'ERROR: The image stack file "{cfg.img_stack}" does not exist')
            sys.exit(3)

        fiji_app = smlfm.FijiApp()
        if not fiji_app.has_imagej:
            print('PyImageJ not installed')
        else:
            print('Detecting Fiji...')
            fiji_app.detect_fiji(cfg.fiji_dir, cfg.fiji_jvm_opts)
            if not fiji_app.has_fiji:
                print('Fiji not detected')
            else:
                if cfg.csv_file is None:
                    cfg.csv_file = Path(str(cfg.img_stack) + '.csv')
                cfg.csv_format = smlfm.LocalisationFile.Format.PEAKFIT
                fiji_app.run_peakfit(cfg.img_stack, cfg.csv_file)

    if cfg.csv_file is None:
        print('ERROR: The CSV file not set')
        sys.exit(1)
    if not cfg.csv_file.exists():
        print(f'ERROR: The CSV file "{cfg.csv_file}" does not exist')
        sys.exit(2)

    lfm = smlfm.FourierMicroscope(
        cfg.num_aperture, cfg.mla_lens_pitch,
        cfg.focal_length_mla, cfg.focal_length_obj_lens,
        cfg.focal_length_tube_lens, cfg.focal_length_fourier_lens,
        cfg.pixel_size_camera, cfg.ref_idx_immersion,
        cfg.ref_idx_medium)

    # 2. Read localisation file

    tic = time.time()

    csv = smlfm.LocalisationFile(cfg.csv_file, cfg.csv_format)
    csv.read()

    print(f'Loaded {csv.data.shape[0]} localisations from'
          f' {np.unique(csv.data[:, 0]).shape[0]} unique frames')

    if cfg.show_graphs and cfg.show_all_debug_graphs:
        fig = smlfm.graphs.draw_locs_csv(plt.figure(), csv.data[:, 1:3])
        fig.canvas.manager.set_window_title('Raw localisations')

    csv.scale_peakfit_data(lfm.pixel_size_sample)
    locs_2d_csv = csv.data.copy()

    # Center X and Y to their means
    # TODO: Why center the data to means? Ensure how is this related to MLA centre.
    locs_2d_csv[:, 1] -= locs_2d_csv[:, 1].mean()
    locs_2d_csv[:, 2] -= locs_2d_csv[:, 2].mean()

    if cfg.log_timing:
        print(f'Loading {repr(cfg.csv_file.name)} took {1e3 * (time.time() - tic):.3f} ms')

    # 3. Prepare MLA and rotate it to match the CSV data

    mla = smlfm.MicroLensArray(
        cfg.mla_type, cfg.focal_length_mla, cfg.mla_lens_pitch,
        cfg.mla_optic_size, np.array(cfg.mla_centre))

    if cfg.show_graphs and cfg.show_all_debug_graphs:
        fig = smlfm.graphs.draw_mla(plt.figure(), mla.lens_centres, mla.centre)
        fig.canvas.manager.set_window_title('Micro-lens array centres')

    mla.rotate_lattice(np.deg2rad(cfg.mla_rotation))
    mla.offset_lattice(np.array(cfg.mla_offset) / lfm.mla_to_xy_scale)  # XY -> MLA

    if cfg.show_graphs and cfg.show_all_debug_graphs:
        fig = smlfm.graphs.draw_mla(plt.figure(), mla.lens_centres, mla.centre)
        fig.canvas.manager.set_window_title('Micro-lens array centres rotated')

    # 4. Map localisations to lenses

    tic = time.time()

    lfl = smlfm.Localisations(locs_2d_csv)
    lfl.assign_to_lenses(mla, lfm)

    if cfg.log_timing:
        print(f'Mapping points to lenses took {1e3 * (time.time() - tic):.3f} ms')

    if cfg.show_graphs and cfg.show_mla_alignment_graph:
        fig = smlfm.graphs.draw_locs(
            plt.figure(),
            xy=lfl.locs_2d[:, 3:5],
            lens_idx=lfl.locs_2d[:, 12],
            lens_centres=(mla.lens_centres - mla.centre) * lfm.mla_to_xy_scale,
            # U,V values are around MLA centre but that offset is not included
            mla_centre=np.array([0.0, 0.0]))
        fig.canvas.manager.set_window_title('Localisations with lens centers')

        # Ask the user to confirm if micro-lenses are correctly aligned
        if cfg.confirm_mla_alignment:
            tic = time.time()
            print('')
            print('Check on the figure that the lenses are well aligned with the'
                  ' data. Then close the window(s) to continue.')
            print('If the alignment is incorrect, answer No in followup question,'
                  ' adjust MLA rotation and/or offset in the configuration,'
                  ' and run this application again.')
            plt.show()
            while True:
                data = input('\nAre the lens centres aligned with the data? [Y/n]: ')
                data = 'y' if data == '' else data[0].casefold()
                if data not in ('y', 'n'):
                    print('Not an appropriate choice.')
                else:
                    if data == 'n':
                        sys.exit(10)
                    break
            print('')
            user_interaction_time = time.time() - tic

    # 5. Filter localisations and set alpha model

    tic = time.time()

    if cfg.filter_lenses:
        lfl.filter_lenses(mla, lfm)
    if cfg.filter_rhos is not None:
        lfl.filter_rhos(cfg.filter_rhos)
    if cfg.filter_spot_sizes is not None:
        lfl.filter_spot_sizes(cfg.filter_spot_sizes)
    if cfg.filter_photons is not None:
        lfl.filter_photons(cfg.filter_photons)

    lfl.init_alpha_uv(cfg.alpha_model, lfm, worker_count=cfg.max_workers)

    if cfg.log_timing:
        print(f'Filtering and setting alpha model took {1e3 * (time.time() - tic):.3f} ms')

    # 6. Find system aberrations

    tic = time.time()

    fit_params_cor = dataclasses.replace(
        cfg.fit_params_aberration,
        frame_min=(cfg.fit_params_aberration.frame_min
                   if cfg.fit_params_aberration.frame_min > 0
                   else lfl.min_frame),
        frame_max=(cfg.fit_params_aberration.frame_max
                   if cfg.fit_params_aberration.frame_max > 0
                   else min(1000, lfl.max_frame)),
    )

    print(f'Fitting frames'
          f' {fit_params_cor.frame_min}-{fit_params_cor.frame_max}'
          f' for aberration correction...')
    _, fit_data = smlfm.Fitting.light_field_fit(
        lfl.filtered_locs_2d, lfm.rho_scaling, fit_params_cor,
        worker_count=cfg.max_workers)

    correction = smlfm.Fitting.calculate_view_error(
        lfl.filtered_locs_2d, lfm.rho_scaling, fit_data, cfg.aberration_params)
    # print('Global aberration calculated [\u03BCm] (u,v,dx,dy,count):')
    # with np.printoptions(precision=6, suppress=True):
    #     print(f'{correction}')

    lfl.correct_xy(correction)

    if cfg.show_graphs and cfg.show_all_debug_graphs:
        fig = smlfm.graphs.draw_locs(
            plt.figure(),
            xy=lfl.corrected_locs_2d[:, 3:5],
            lens_idx=lfl.corrected_locs_2d[:, 12],
            lens_centres=(mla.lens_centres - mla.centre) * lfm.mla_to_xy_scale,
            # U,V values are around MLA centre but that offset is not included
            mla_centre=np.array([0.0, 0.0]))
        fig.canvas.manager.set_window_title('Corrected localisations')

    if cfg.log_timing:
        print(f'Aberration correction took {1e3 * (time.time() - tic):.3f} ms')

    # 7. Fit full data set on corrected localisations

    tic = time.time()

    fit_params_all = dataclasses.replace(
        cfg.fit_params_full,
        frame_min=(cfg.fit_params_full.frame_min
                   if cfg.fit_params_full.frame_min > 0
                   else lfl.min_frame),
        frame_max=(cfg.fit_params_full.frame_max
                   if cfg.fit_params_full.frame_max > 0
                   else lfl.max_frame),
    )

    print(f'Fitting frames'
          f' {fit_params_all.frame_min}-{fit_params_all.frame_max}...')
    locs_3d, _ = smlfm.Fitting.light_field_fit(
        lfl.corrected_locs_2d, lfm.rho_scaling, fit_params_all,
        worker_count=cfg.max_workers,
        progress_func=lambda frame, min_frame, max_frame:
            print(f'Processing frame'
                  f' {frame - min_frame + 1}/{max_frame - min_frame + 1}...'))

    print(f'Total number of frames used for fitting:'
          f' {np.unique(locs_3d[:, 7]).shape[0]}')
    print(f'Total number of 2D localisations used for fitting:'
          f' {int(np.sum(locs_3d[:, 5]))}')
    print(f'Total number of 3D localisations: {locs_3d.shape[0]}')

    if cfg.log_timing:
        print(f'Complete fitting took {1e3 * (time.time() - tic):.3f} ms')

    # 8. Write the results

    if cfg.save_dir is not None and cfg.save_dir:

        timestamp_str = timestamp.strftime('%Y%m%d-%H%M%S')
        subdir_name = Path(f'3D Fitting - {timestamp_str} - {cfg.csv_file.name}')

        results = smlfm.OutputFiles(cfg, subdir_name)
        print(f'Saving results to folder: "{results.folder.name}"')

        try:
            results.mkdir()
        except Exception as ex:
            print(f'ERROR: Failed to create target folder ({repr(ex)})')
        else:
            try:
                results.save_config()
            except Exception as ex:
                print(f'ERROR: Failed to save configuration file ({repr(ex)})')
            try:
                results.save_csv(locs_3d)
            except Exception as ex:
                print(f'ERROR: Failed to save CSV file ({repr(ex)})')
            try:
                results.save_visp(locs_3d)
            except Exception as ex:
                print(f'ERROR: Failed to save ViSP file ({repr(ex)})')
            try:
                results.save_figures()
            except Exception as ex:
                print(f'ERROR: Failed to save figures file ({repr(ex)})')

    # 9. Plotting results

    if cfg.show_graphs and cfg.show_result_graphs:
        fig1, fig2, fig3 = smlfm.graphs.reconstruct_results(
            plt.figure(), plt.figure(), plt.figure(),
            locs_3d, cfg.show_max_lateral_err, cfg.show_min_view_count)
        fig1.canvas.manager.set_window_title('Occurrences')
        fig2.canvas.manager.set_window_title('Histogram')
        fig3.canvas.manager.set_window_title('3D')

    # End

    if cfg.log_timing:
        total_time = time.time() - tic_total - user_interaction_time
        print(f'Total time: {1e3 * total_time:.3f} ms')

    if cfg.show_graphs:
        plt.show()


if __name__ == '__main__':
    app()



import json
from dataclasses import asdict, dataclass, field, fields, is_dataclass
from enum import Enum
from pathlib import Path, PurePath
from typing import Optional, Tuple

from .fitting import Fitting
from .localisation_file import LocalisationFile
from .localisations import Localisations
from .micro_lens_array import MicroLensArray


@dataclass
# pylint: disable=too-many-instance-attributes
class Config:
    _fiji_dir_doc: str = (
        'A path to Fiji.app folder with Fiji application.\n'
        'Fiji is needed to find localizations in an image stack via PeakFit plugin.\n'
        'If set to \'null\' or empty string, a CSV file must be given.')
    fiji_dir: Optional[Path] = None
    _fiji_jvm_opts_doc: str = (
        'A list of space separated options for Java Virtual Machine started by Fiji.')
    fiji_jvm_opts: str = '-Xmx10g'
    _img_stack_doc: str = (
        'A path to image stack file to be analyzed by PeakFit plugin in Fiji.\n'
        'If the Fiji application is not detected, this value is ignored.\n'
        'Otherwise, the output from PeakFit plugin is stored in configured CSV file.')
    img_stack: Optional[Path] = None
    _csv_file_doc: str = (
        'A path to CSV/XLS file with localisations.\n'
        'If the image stack file is given and if Fiji application is found,\n'
        'this file is overwritten with PeakFit output.')
    csv_file: Optional[Path] = None
    _csv_format_doc: str = (
        'A format of CSV file with localisations.\n'
        'Supported values: \'PEAKFIT\', \'THUNDERSTORM\', \'PICASSO\'.')
    csv_format: LocalisationFile.Format = LocalisationFile.Format.PEAKFIT

    _mla_type_doc: str = (
        'A Micro-Lens Array lens arrangement.\n'
        'Supported values: \'HEXAGONAL\', \'SQUARE\'.')
    mla_type: MicroLensArray.LatticeType = MicroLensArray.LatticeType.HEXAGONAL
    _mla_lens_pitch_doc: str = (
        'A pitch along X direction (in microns).')
    mla_lens_pitch: float = 2390.0
    _mla_optic_size_doc: str = (
        'A size of MLA optic (in microns).')
    mla_optic_size: float = 10000.0
    _mla_centre_doc: str = (
        'An X,Y coordinates of the array center (in lattice spacing units).\n'
        'The MLA keeps the distance between lens centres equal to 1.')
    mla_centre: Tuple[float, float] = field(default_factory=lambda: (0.0, 0.0))
    _mla_rotation_doc: str = (
        'A rotation of the MLA to match the orientation of the data (in degrees)')
    mla_rotation: float = 30.8
    _mla_offset_doc: str = (
        'An X,Y offset of the MLA to match the orientation of the data (in microns).')
    mla_offset: Tuple[float, float] = field(default_factory=lambda: (0.0, 0.0))
    _focal_length_mla_doc: str = (
        'A focal length of Micro-Lens Array (in mm).')
    focal_length_mla: float = 175.0

    _focal_length_obj_lens_doc: str = (
        'A focal length of the objective lens (in mm).')
    focal_length_obj_lens: float = 200.0 / 60.0
    _focal_length_fourier_lens_doc: str = (
        'A focal length of the fourier lens (in mm).')
    focal_length_fourier_lens: float = 175.0
    _focal_length_tube_lens_doc: str = (
        'A focal length of the tube lens (in mm).')
    focal_length_tube_lens: float = 200.0

    _num_aperture_doc: str = (
        'A numerical aperture of the objective.')
    num_aperture: float = 1.27
    _ref_idx_immersion_doc: str = (
        'An immersion refractive index.')
    ref_idx_immersion: float = 1.33
    _ref_idx_medium_doc: str = (
        'A specimen/medium refractive index.')
    ref_idx_medium: float = 1.33

    _pixel_size_camera_doc: str = (
        'A camera pixel size (in microns).')
    pixel_size_camera: float = 16.0

    _filter_lenses_doc: str = (
        'Filter out all localizations mapped to the lenses with center\n'
        'lying outside the back focal plane.')
    filter_lenses: bool = True
    _filter_rhos_doc: str = (
        'Filter out all localizations with distance to origin\n'
        'lying outside given min,max range (in normalised pupil coordinates).')
    filter_rhos: Optional[Tuple[float, float]] = None
    _filter_spot_sizes_doc: str = (
        'Filter out all localizations with spot size\n'
        'lying outside given min,max range (in microns).')
    filter_spot_sizes: Optional[Tuple[float, float]] = None
    _filter_photons_doc: str = (
        'Filter out all localizations with background intensity\n'
        'lying outside given min,max range (in photons).')
    filter_photons: Optional[Tuple[float, float]] = None
    _alpha_model_doc: str = (
        'A model used for mapping.\n'
        'Supported values: \'LINEAR\', \'SPHERE\', \'INTEGRATE_SPHERE\'.')
    alpha_model: Localisations.AlphaModel = Localisations.AlphaModel.INTEGRATE_SPHERE

    _fit_params_aberration_doc: str = (
        'A fitting parameters for aberration correction purposes.')
    _fit_params_aberration_frame_min_doc: str = (
        'A min. frame number, zero or negative to auto-fill min. value.')
    _fit_params_aberration_frame_max_doc: str = (
        'A max. frame number, zero or negative to auto-fill default value 1000.')
    _fit_params_aberration_disparity_max_doc: str = (
        'A limit of possible disparity between adjacent views,\n'
        'from negative to positive value (in microns).')
    _fit_params_aberration_disparity_step_doc: str = (
        'A step used for disparity range (in microns).')
    _fit_params_aberration_dist_search_doc: str = (
        'A window to accept points as belonging to a given Z (in microns).')
    _fit_params_aberration_angle_tolerance_doc: str = (
        'A tolerance for selecting with angles around relative U,V (in degrees).')
    _fit_params_aberration_threshold_doc: str = (
        'A max. distance of the candidate (in microns).')
    _fit_params_aberration_min_views_doc: str = (
        'A min. number of views the candidate is seen to count it.')
    _fit_params_aberration_z_calib_doc: str = (
        'A calibration factor between optical and physical Z axis.\n'
        'Unused and ignored for aberration correction.')
    fit_params_aberration: Fitting.FitParams = field(
        default_factory=lambda: Fitting.FitParams(
            frame_min=-1,
            frame_max=-1,
            disparity_max=5.0,
            disparity_step=0.1,
            dist_search=0.5,
            angle_tolerance=2.0,
            threshold=1.0,
            min_views=3,
            z_calib=None,
        )
    )
    _aberration_params_doc: str = (
        'Parameters for aberration correction.')
    _aberration_params_axial_window_doc: str = (
        'A max. standard error in Z to apply correction (in microns).')
    _aberration_params_photon_threshold_doc: str = (
        'A min. number of photons to apply correction.')
    _aberration_params_min_views_doc: str = (
        'A min. number of views the candidate is seen to count it.')
    aberration_params: Fitting.AberrationParams = field(
        default_factory=lambda: Fitting.AberrationParams(
            axial_window=1.0,
            photon_threshold=1,
            min_views=3,
        )
    )
    _fit_params_full_doc: str = (
        'A fitting parameters for full data set, used after aberration correction.')
    _fit_params_full_frame_min_doc: str = (
        'A min. frame number, zero or negative to auto-fill min. value.')
    _fit_params_full_frame_max_doc: str = (
        'A max. frame number, zero or negative to auto-fill max. value.')
    _fit_params_full_disparity_max_doc: str = (
        'A limit of possible disparity between adjacent views,\n'
        'from negative to positive value (in microns).')
    _fit_params_full_disparity_step_doc: str = (
        'A step used for disparity range (in microns).')
    _fit_params_full_dist_search_doc: str = (
        'A window to accept points as belonging to a given Z (in microns).')
    _fit_params_full_angle_tolerance_doc: str = (
        'A tolerance for selecting with angles around relative U,V (in degrees).')
    _fit_params_full_threshold_doc: str = (
        'A max. distance of the candidate (in microns).')
    _fit_params_full_min_views_doc: str = (
        'A min. number of views the candidate is seen to count it.')
    _fit_params_full_z_calib_doc: str = (
        'A calibration factor between optical and physical Z axis.')
    fit_params_full: Fitting.FitParams = field(
        default_factory=lambda: Fitting.FitParams(
            frame_min=-1,
            frame_max=-1,
            disparity_max=8.0,
            disparity_step=0.1,
            dist_search=0.5,
            angle_tolerance=1.0,
            threshold=0.3,
            min_views=2,
            z_calib=1.53,
        )
    )

    _save_dir_doc: str = (
        'Save the results to a new sub-folder under this directory.\n'
        'If set to \'null\' or empty string, nothing is saved.\n'
        'Defaults to current working directory.')
    save_dir: Optional[Path] = Path()

    _show_graphs_doc: str = (
        'Show various graphs, a global switch.\n'
        'In CLI app it pops up GUI windows, so e.g. on Linux an X server\n'
        'or Wayland must be running and DISPLAY env. var. must be set.')
    show_graphs: bool = True
    _show_result_graphs_doc: str = (
        'Show only final three graphs with 3D view, histogram and errors.')
    show_result_graphs: bool = True
    _show_mla_alignment_graph_doc: str = (
        'Show micro-lenses together with localisations.')
    show_mla_alignment_graph: bool = True
    _show_all_debug_graphs_doc: str = (
        'Show all remaining graphs.')
    show_all_debug_graphs: bool = False

    _show_max_lateral_err_doc: str = (
        'In result graphs show only points with lateral error below this value.')
    show_max_lateral_err: Optional[float] = None
    _show_min_view_count_doc: str = (
        'In result graphs show only points with view count below this value.')
    show_min_view_count: Optional[int] = None

    _confirm_mla_alignment_doc: str = (
        'Ask the user to confirm the data and MLA alignment.\n'
        'This is only active with \'show_mla_alignment_graph\'.')
    confirm_mla_alignment: bool = True

    _log_timing_doc: str = (
        'Print timing messages to console for all lengthy operations.')
    log_timing: bool = True

    _max_workers_doc: str = (
        'Limit the number of parallel workers used to speed up some computations.\n'
        'If set to zero or less all CPUs in the system are used.')
    max_workers: int = 0

    def to_json(self, **kwargs):
        indent = kwargs.pop('indent', 4)  # Indent with 4 spaces by default

        class JSONEncoder(json.JSONEncoder):
            def default(self, o):
                if is_dataclass(o):
                    return asdict(o)
                if isinstance(o, PurePath):
                    return str(o)
                if isinstance(o, Enum):
                    return o.name
                return super().default(o)

        return json.dumps(self, cls=JSONEncoder,
                          indent=indent, ensure_ascii=False,
                          **kwargs)

    @staticmethod
    def from_json(dump: str):
        cfg = Config(**json.loads(dump))
        cfg_default = Config()

        for f in fields(Config):
            # Override doc fields with internal documentation
            if f.name.startswith('_') and f.name.endswith('_doc'):
                setattr(cfg, f.name, getattr(cfg_default, f.name))
                continue

            value = getattr(cfg, f.name)
            if repr(f.type) == repr(Optional[Path]):
                if value is not None:
                    setattr(cfg, f.name, Path(value))
            elif repr(f.type) == repr(Tuple[float, float]):
                setattr(cfg, f.name, tuple(value))
            elif repr(f.type) == repr(Optional[Tuple[float, float]]):
                if value is not None:
                    setattr(cfg, f.name, tuple(value))
            elif f.type is LocalisationFile.Format:
                setattr(cfg, f.name, LocalisationFile.Format[value])
            elif f.type is Localisations.AlphaModel:
                setattr(cfg, f.name, Localisations.AlphaModel[value])
            elif f.type is MicroLensArray.LatticeType:
                setattr(cfg, f.name, MicroLensArray.LatticeType[value])
            elif f.type is Fitting.FitParams:
                setattr(cfg, f.name, Fitting.FitParams(**value))
            elif f.type is Fitting.AberrationParams:
                setattr(cfg, f.name, Fitting.AberrationParams(**value))

        return cfg



import numpy.typing as npt
from matplotlib.figure import Figure

from ..graphs import add_watermark


def draw_occurrences(fig: Figure,
                     lateral_err: npt.NDArray[float],
                     axial_err: npt.NDArray[float],
                     photons: npt.NDArray[float],
                     bins: int = 40,
                     set_default_size: bool = True
                     ) -> Figure:
    fig.clear(True)
    if set_default_size:
        fig.set_size_inches(15, 5)
    fig.set_layout_engine('tight')

    ax0, ax1, ax2 = fig.subplots(ncols=3)
    ax0.set_xlabel('Lateral fit error [nm]')
    ax0.set_ylabel('Occurrence')
    ax0.hist(lateral_err * 1000, bins=bins)

    ax1.set_xlabel('Axial fit error [nm]')
    ax1.set_ylabel('Occurrence')
    ax1.hist(axial_err * 1000, bins=bins)

    ax2.set_xlabel('Number of photons')
    ax2.set_ylabel('Occurrence')
    ax2.hist(photons, bins=bins)

    add_watermark(fig)
    return fig



class FourierMicroscope:
    """A class describing fourier light field microscope.

    Attributes:
        num_aperture (float): A numerical aperture of the objective.
        mla_lens_pitch (float):
            A pitch along X direction, parallel with optical table (in microns).
        focal_length_mla (float):
            A focal length of micro-lens array (in mm).
        focal_length_obj_lens (float):
            A focal length of the objective lens (in mm).
        focal_length_tube_lens (float):
            A focal length of the tube lens (in mm).
        focal_length_fourier_lens (float):
            A focal length of the fourier lens (in mm).
        pixel_size_camera (float): A camera pixel size (in microns).
        ref_idx_immersion (float): An immersion refractive index.
        ref_idx_medium (float): A specimen/medium refractive index.

        bfp_radius (float):
            A radius of the conjugate back focal plane (in microns).
        bfp_lens_count (float):
            A number of micro-lenses in the conjugate back focal plane.
        pixels_per_lens (float): A number of camera pixels per micro-lens.
        magnification (float):
            An overall magnification to camera plane, from MLA lattice spacing
            units to microns.
        pixel_size_sample (float): A pixel size in sample space (in microns).
        rho_scaling (float):
            A scaling factor to convert microns in image plane to rho.

        xy_to_uv_scale (float):
            A scale factor to convert X,Y coordinates to U,V units.
        mla_to_uv_scale (float):
            A scale factor to convert MLA lens centers coordinates to U,V units.
        mla_to_xy_scale (float):
            A scale factor to convert MLA lens centers coordinates to X,Y units.
    """

    def __init__(self,
                 num_aperture: float,
                 mla_lens_pitch: float,
                 focal_length_mla: float,
                 focal_length_obj_lens: float,
                 focal_length_tube_lens: float,
                 focal_length_fourier_lens: float,
                 pixel_size_camera: float,
                 ref_idx_immersion: float,
                 ref_idx_medium: float):
        """Constructs fourier light field microscope object.

        Args:
            num_aperture (float): A numerical aperture of the objective.
            mla_lens_pitch (float):
                A pitch along X direction, parallel with optical table
                (in microns).
            focal_length_mla (float):
                A focal length of micro-lens array (in mm).
            focal_length_obj_lens (float):
                A focal length of the objective lens (in mm).
            focal_length_tube_lens (float):
                A focal length of the tube lens (in mm).
            focal_length_fourier_lens (float):
                A focal length of the fourier lens (in mm).
            pixel_size_camera (float): A camera pixel size (in microns).
            ref_idx_immersion (float): An immersion refractive index.
            ref_idx_medium (float): A specimen/medium refractive index.
        """

        self.num_aperture = num_aperture
        self.mla_lens_pitch = mla_lens_pitch
        self.focal_length_mla = focal_length_mla
        self.focal_length_obj_lens = focal_length_obj_lens
        self.focal_length_tube_lens = focal_length_tube_lens
        self.focal_length_fourier_lens = focal_length_fourier_lens
        self.pixel_size_camera = pixel_size_camera
        self.ref_idx_immersion = ref_idx_immersion
        self.ref_idx_medium = ref_idx_medium

        self.bfp_radius = (
                1000 * num_aperture * focal_length_obj_lens
                * (focal_length_fourier_lens / focal_length_tube_lens))
        self.bfp_lens_count = 2.0 * self.bfp_radius / mla_lens_pitch
        self.pixels_per_lens = mla_lens_pitch / pixel_size_camera
        self.magnification = (
                (focal_length_tube_lens / focal_length_obj_lens)
                * (focal_length_mla / focal_length_fourier_lens))
        self.pixel_size_sample = pixel_size_camera / self.magnification
        self.rho_scaling = self.magnification / self.bfp_radius

        self.xy_to_uv_scale = self.rho_scaling
        self.mla_to_uv_scale = 2.0 / self.bfp_lens_count
        self.mla_to_xy_scale = self.mla_to_uv_scale / self.xy_to_uv_scale



import multiprocessing as mp
import warnings
from enum import Enum, unique
from multiprocessing.shared_memory import SharedMemory as mp_SharedMemory
from typing import Optional, Tuple

import numpy as np
import numpy.typing as npt
from sklearn.neighbors import NearestNeighbors

from .fourier_microscope import FourierMicroscope
from .micro_lens_array import MicroLensArray


class Localisations:
    """A class containing light field localisation data.

    Attributes:
        locs_2d (npt.NDArray[float]): A localisation data in rows, where each
            has 13 columns in this order:

            * [0] frame number (int)
            * [1] U coordinate (in normalised pupil coordinates)
            * [2] V coordinate (in normalised pupil coordinates)
            * [3] X coordinate (in microns)
            * [4] Y coordinate (in microns)
            * [5] sigma X (in microns)
            * [6] sigma Y (in microns)
            * [7] intensity (in photons)
            * [8] background intensity (in photons)
            * [9] precision or uncertainty (in microns)
            * [10] alpha U (in normalised pupil coordinates)
            * [11] alpha V (in normalised pupil coordinates)
            * [12] lens index (int)

        min_frame (int): The smallest frame number in the data.
        max_frame (int): The largest frame number in the data.
        filtered_locs_2d (npt.NDArray[float] or None):
            A filtered localisations from `locs_2d` attribute,
            or None if not set via `init_alpha_uv()` function yet.
        alpha_model (AlphaModel or None):
            A model used for mapping,
            or None if not set via `init_alpha_uv()` function yet.
        corrected_locs_2d (npt.NDArray[float] or None):
            A corrected localisations from `filtered_locs_2d` attribute,
            or None if not set via `correct_xy()` function yet.
        correction (npt.NDArray[float] or None):
            An average fit error across all fitted points (in microns),
            or None if not set via `correct_xy()` function yet.
    """

    @unique
    class AlphaModel(Enum):
        LINEAR = 1
        SPHERE = 2
        INTEGRATE_SPHERE = 3

    def __init__(self,
                 locs_2d_csv: npt.NDArray[float]
                 ):
        """Constructs light field localisation data object.

        Args:
            locs_2d_csv (npt.NDArray[float]):
                A localisation data read from localisation file.
        """

        self.min_frame = int(np.min(locs_2d_csv[:, 0]))
        self.max_frame = int(np.max(locs_2d_csv[:, 0]))

        self.locs_2d = np.zeros((locs_2d_csv.shape[0], 13))

        self.locs_2d[:, 0] = locs_2d_csv[:, 0].copy()
        self.locs_2d[:, 3:10] = locs_2d_csv[:, 1:8].copy()
        # The locs_2d columns 1, 2, 12 are initialized in assign_to_lenses()
        # The locs_2d columns 10, 11 remain zeroed
        # The filtered_locs_2d columns 10, 11 are initialized in init_alpha_uv()

        self.filtered_locs_2d: Optional[npt.NDArray[float]] = None
        self.alpha_model: Optional[Localisations.AlphaModel] = None
        self.corrected_locs_2d: Optional[npt.NDArray[float]] = None
        self.correction: Optional[npt.NDArray[float]] = None

    def assign_to_lenses(self,
                         mla: MicroLensArray,
                         lfm: FourierMicroscope
                         ) -> None:
        """Initialize U,V values from X,Y mapped to MLA lenses.

        Args:
            mla (MicroLensArray): An instance of micro-lens array class.
            lfm (FourierMicroscope):
                An instance of fourier light field microscope class.
        """

        xy = self.locs_2d[:, 3:5].copy()
        xy *= lfm.xy_to_uv_scale

        lens_centres = mla.lens_centres - mla.centre
        lens_centres *= lfm.mla_to_uv_scale

        knn = NearestNeighbors(n_neighbors=1).fit(lens_centres)
        lens_indices = knn.kneighbors(xy, return_distance=False)

        self.locs_2d[:, 1:3] = lens_centres[lens_indices, :][:, 0, :]  # U, V
        self.locs_2d[:, 12] = lens_indices[:, 0]

        self.reset_filtered_locs()

    def _filter(self,
                filter_range: Tuple[float, float],
                column_data: npt.NDArray[float]
                ) -> None:
        sel = np.logical_or(column_data < filter_range[0],
                            column_data > filter_range[1])
        self.filtered_locs_2d = self.filtered_locs_2d[~sel, :]

    def reset_filtered_locs(self) -> None:
        self.filtered_locs_2d = self.locs_2d.copy()
        self.alpha_model = None
        self.reset_correction()

    def filter_lenses(self,
                      mla: MicroLensArray,
                      lfm: FourierMicroscope
                      ) -> None:
        radius = lfm.bfp_radius / mla.lens_pitch
        radius_sq = radius ** 2
        centres = mla.lens_centres - mla.centre
        distance_sq = np.sum(centres ** 2, axis=1)
        lens_indices = np.nonzero(distance_sq < radius_sq)[0]

        index = self.filtered_locs_2d[:, 12]
        sel = np.any(index[:, None] == lens_indices, axis=-1)
        self.filtered_locs_2d = self.filtered_locs_2d[sel, :]

    def filter_rhos(self,
                    filter_range: Tuple[float, float]
                    ) -> None:
        uv = self.filtered_locs_2d[:, 1:3]
        rhos = np.sqrt(np.sum(uv ** 2, axis=1))
        self._filter(filter_range, rhos)

    def filter_spot_sizes(self,
                          filter_range: Tuple[float, float]
                          ) -> None:
        spot_sizes = self.filtered_locs_2d[:, 5]
        self._filter(filter_range, spot_sizes)

    def filter_photons(self,
                       filter_range: Tuple[float, float]
                       ) -> None:
        photons = self.filtered_locs_2d[:, 7]
        self._filter(filter_range, photons)

    def init_alpha_uv(self,
                      model: AlphaModel,
                      lfm: FourierMicroscope,
                      abort_event: Optional[mp.Event] = None,
                      worker_count: int = 0
                      ) -> None:
        """TODO: Add documentation.

        Args:
            model (AlphaModel): A model used for mapping.
            lfm (FourierMicroscope):
                An instance of fourier light field microscope class.
            abort_event (Optional[mp.Event]):
                An event to be periodically checked (without blocking).
                If set, the processing is aborted and function returns.
            worker_count (int):
                It is the number of worker processes to run in parallel.
                If it is 0 or less, the number returned by `os.cpu_count()` is used.
        """

        uv = self.filtered_locs_2d[:, 1:3]
        alpha_uv = self.filtered_locs_2d[:, 10:12]
        na = lfm.num_aperture
        n = lfm.ref_idx_medium

        self.reset_correction()

        if model == Localisations.AlphaModel.LINEAR:
            alpha_uv[:] = uv

        elif model == Localisations.AlphaModel.SPHERE:
            rho = np.sqrt(np.sum(uv**2, axis=1))
            dr_sq = 1 - rho * (na / n)**2  # TODO: REVIEW! Should we use (rho**2) here?
            dr_sq[dr_sq < 0.0] = np.nan  # No negative numbers for sqrt
            phi = -(na / n) / np.sqrt(dr_sq)
            alpha_uv[:] = uv * phi[:, np.newaxis]

        elif model == Localisations.AlphaModel.INTEGRATE_SPHERE:
            uv_scaling = lfm.mla_to_uv_scale
            m = 10
            self._phase_average_sphere(
                uv, uv_scaling, na, n, m, alpha_uv, abort_event, worker_count)
            if abort_event is not None:
                if abort_event.is_set():
                    alpha_uv[:] = 0
                    self.alpha_model = None
                    return

        else:
            raise ValueError(f'Unsupported alpha model with value {model}')

        self.alpha_model = model

    def reset_correction(self) -> None:
        self.corrected_locs_2d = None
        self.correction = None

    def correct_xy(self,
                   correction: npt.NDArray[float]
                   ) -> None:
        self.corrected_locs_2d = self.filtered_locs_2d.copy()
        self.correction = correction.copy()

        # These variables are not modified, no need to copy the data
        u = self.corrected_locs_2d[:, 1]
        v = self.corrected_locs_2d[:, 2]
        # Through these variables are modified the original data
        x = self.corrected_locs_2d[:, 3]
        y = self.corrected_locs_2d[:, 4]

        for i in range(correction.shape[0]):
            idx = np.logical_and(u == correction[i, 0], v == correction[i, 1])
            x[idx] -= correction[i, 2]
            y[idx] -= correction[i, 3]

    _SHM_NAME__PHASE_AVG__UV = 'smlfm-phase_avg_sphere-uv'
    _SHM_NAME__PHASE_AVG__ALPHA_UV = 'smlfm-phase_avg_sphere-alpha_uv'

    @staticmethod
    def _phase_average_sphere(uv: npt.NDArray[float],
                              uv_scaling: float,
                              na: float,
                              n: float,
                              m: int,
                              alpha_uv: npt.NDArray[float],
                              abort_event: Optional[mp.Event] = None,
                              worker_count: int = 0
                              ) -> None:
        """TODO: Add documentation."""

        rows_per_task = 500
        row_count = uv.shape[0]
        task_count = int((row_count + rows_per_task - 1) / rows_per_task)

        if worker_count == 1 or task_count == 1:
            Localisations._phase_average_sphere_fn(
                range(row_count), uv, uv_scaling, na, n, m,
                alpha_uv, abort_event)
            return

        shm_bytes = uv.shape[0] * uv.shape[1] * uv.itemsize

        shm_uv = mp_SharedMemory(
            Localisations._SHM_NAME__PHASE_AVG__UV,
            create=True, size=shm_bytes)
        shm_alpha_uv = mp_SharedMemory(
            Localisations._SHM_NAME__PHASE_AVG__ALPHA_UV,
            create=True, size=shm_bytes)

        uv_data = np.ndarray(
            shape=uv.shape, dtype=float, buffer=shm_uv.buf)
        alpha_uv_data = np.ndarray(
            shape=uv.shape, dtype=float, buffer=shm_alpha_uv.buf)

        # Fill shared memory with input data
        uv_data[:] = uv

        try:
            processes = worker_count if worker_count > 0 else None
            with mp.Pool(processes=processes) as pool:
                procs = [Optional[mp.Process]] * task_count

                for idx in range(task_count):
                    i_range_n = range(rows_per_task * idx,
                                      min(rows_per_task * (idx + 1), row_count))
                    # Cannot pass the `abort_event` to processes, otherwise following
                    # runtime error is raised: "Conditional objects should only be
                    # shared between processes through inheritance"
                    args = (i_range_n, row_count, uv_scaling, na, n, m, None)
                    procs[idx] = pool.apply_async(
                        Localisations._phase_average_sphere_task, args)

                for idx in range(task_count):
                    procs[idx].get()
                    if abort_event is not None:
                        if abort_event.is_set():
                            return

        finally:
            if abort_event is None or not abort_event.is_set():
                # Copy output data to target location
                alpha_uv[:] = alpha_uv_data
            shm_alpha_uv.unlink()
            shm_uv.unlink()

    @staticmethod
    def _phase_average_sphere_task(i_range: range,
                                   uv_rows: int,
                                   uv_scaling: float,
                                   na: float,
                                   n: float,
                                   m: int,
                                   abort_event: Optional[mp.Event] = None
                                   ) -> None:
        shm_uv = mp_SharedMemory(
            Localisations._SHM_NAME__PHASE_AVG__UV)
        shm_alpha_uv = mp_SharedMemory(
            Localisations._SHM_NAME__PHASE_AVG__ALPHA_UV)
        try:
            uv = np.ndarray(
                shape=(uv_rows, 2), dtype=float, buffer=shm_uv.buf)
            alpha_uv = np.ndarray(
                shape=(uv_rows, 2), dtype=float, buffer=shm_alpha_uv.buf)

            Localisations._phase_average_sphere_fn(
                i_range, uv, uv_scaling, na, n, m, alpha_uv, abort_event)
        finally:
            shm_alpha_uv.close()
            shm_uv.close()

    @staticmethod
    def _phase_average_sphere_fn(i_range: range,
                                 uv: npt.NDArray[float],
                                 uv_scaling: float,
                                 na: float,
                                 n: float,
                                 m: int,
                                 alpha_uv: npt.NDArray[float],
                                 abort_event: Optional[mp.Event] = None
                                 ) -> None:
        ds2 = uv_scaling / 2

        for i in i_range:
            if abort_event is not None:
                if abort_event.is_set():
                    return

            range_u = np.linspace(uv[i, 0] - ds2, uv[i, 0] + ds2, m + 1)
            range_v = np.linspace(uv[i, 1] - ds2, uv[i, 1] + ds2, m + 1)
            um, vm = np.meshgrid(range_u, range_v)

            rho_sq = um**2 + vm**2
            dr_sq = 1 - rho_sq * (na / n)**2
            dr_sq[dr_sq < 0.0] = np.nan  # No negative numbers for sqrt
            phi = -(na / n) / np.sqrt(dr_sq)

            # Suppress RuntimeWarnings if all values in um/vm are NaNs
            with warnings.catch_warnings():
                warnings.simplefilter("ignore", category=RuntimeWarning)
                alpha_uv[i, 0] = np.nanmean(um * phi)
                alpha_uv[i, 1] = np.nanmean(vm * phi)



#!/usr/bin/env python3
import tkinter as tk

from .app_model import AppModel
from .consts import StageType


class App(tk.Tk):
    def __init__(self):
        super().__init__()

        # Remain invisible until the UI is fully constructed
        self.withdraw()

        self._model = AppModel()

        self.title('Single Molecule Light Field Microscopy')
        self.iconphoto(True, self._model.icons.app)

        self._model.create_gui(self)

        # Set default size and position (enlarges to min. size if necessary)
        self.geometry('450x500+0+0')

        # Now it is ready to become visible
        self.deiconify()

        self._model.stage_start_update(StageType.CONFIG)


def app():
    a = App()
    a.mainloop()


if __name__ == '__main__':
    app()



from enum import unique, Enum

COMMAND = 'command'
READONLY = 'readonly'
TEXT = 'text'
TEXTVARIABLE = 'textvariable'
VARIABLE = 'variable'
IMAGE = 'image'


@unique
class GraphType(Enum):
    MLA = 1  # All generated MLA lenses
    CSV_RAW = 2  # All points as loaded from CSV file
    MAPPED = 3  # Colored CSV points as mapped to MLA lenses
    FILTERED = 4  # Same as MAPPED but with some points filtered out
    CORRECTED = 5  # Similar to FILTERED but with aberration corrections
    LOCS_3D = 6  # 3D view
    HISTOGRAM = 7  # 2D histogram
    OCCURRENCES = 8


@unique
class StageType(Enum):
    CONFIG = 1
    CSV = 2
    OPTICS = 3
    FILTER = 4
    CORRECT = 5
    FIT = 6
